include Error.Loc
include Id
include Util

(* 0-tuple = unit, 1-tuple doesn't exist *)

type is_meta = IsMeta | IsUser | IsShare


module Scope = struct
  type scope = {
    types : (user_id * id) list;
    vars : (user_id * id) list;
    conses : (user_id * id) list;
    tyvars : (user_id * id) list
  }

  let empty_scope = { vars = []; conses = []; tyvars = []; types = [] }

  type namespace = NsVar | NsCons | NsTyVar | NsType

  let resolve scope ns = function
    | FromId i -> i
    | UserId s as x ->
       let sc = match ns with NsVar -> scope.vars | NsCons -> scope.conses | NsTyVar -> scope.tyvars | NsType -> scope.types in
       try
         List.assoc x sc
       with
         Not_found -> (Error.failwith "Unbound identifier %s" s)

  let resolve_l loc scope ns = function
    | FromId i -> i
    | UserId s as x ->
       let sc = match ns with NsVar -> scope.vars | NsCons -> scope.conses | NsTyVar -> scope.tyvars | NsType -> scope.types in
       try
         List.assoc x sc
       with
         Not_found -> Error.error loc "Unbound identifier %s@." s

  let intro scope ns = function
    | FromId i -> (i, scope)
    | UserId s as x ->
       let x' = good_id s in
       (x',match ns with
           | NsVar -> {scope with vars = (x,x')::scope.vars}
           | NsCons -> {scope with conses = (x,x')::scope.conses}
           | NsTyVar -> {scope with tyvars = (x,x')::scope.tyvars}
           | NsType -> {scope with types = (x,x')::scope.types})

  let intros scope ns read write l =
    List.fold_left (fun (bindings,scope) b ->
                    let (x,scope) = intro scope ns (read b) in
                    (write b x::bindings,scope)
                   ) ([],scope) (List.rev l)

  let intro_named scope ns x x' =
    match ns with
        | NsVar -> {scope with vars = (x,x')::scope.vars}
        | NsCons -> {scope with conses = (x,x')::scope.conses}
        | NsTyVar -> {scope with tyvars = (x,x')::scope.tyvars}
        | NsType -> {scope with types = (x,x')::scope.types}
end

(* IMPROVE: use GADT for representation
type ('a,'expr) has_match = Has_match
type is_ml = Is_ML

type has_meta = Has_meta
type no_meta = No_meta


type ('id,_) ty_aux =
  | TyArrow : ('id,'match) ty_aux * ('id,'match) ty_aux -> ('id, 'match) ty_aux
  | TyCon : 'id * ('id,'match) ty_aux list -> ('id,'match) ty_aux
  | TyVar : 'id -> ('id,'match) ty_aux
  | TyProduct : ('id, 'match) ty_aux list -> ('id, 'match) ty_aux
  | TyMatch : 'expr -> ('id, ('meta,'expr) has_match)
 *)


type ('id,'expr) ty_aux =
  | TyArrow of ('id,'expr) ty_aux * ('id,'expr) ty_aux
  | TyCon of 'id * ('id,'expr) ty_aux list
  | TyVar of 'id
  | TyProduct of ('id,'expr) ty_aux list
(*
  | TyMatch of 'expr * ('id * 'id list * ('id, 'expr) ty_aux) list
  | TyTyAbs of 'id * ('id,'expr) ty_aux (* Kind ommitted *)
  | TyTyApp of ('id,'expr) ty_aux * ('id,'expr) ty_aux
  | TyTeAbs of 'id * ('id, 'expr) ty_aux * ('id, 'expr) ty_aux
  | TyTeApp of ('id, 'expr) ty_aux * 'expr
 *)

type ty = (id, unit) ty_aux
type user_ty = (user_id,unit) ty_aux

module Implicit = struct
  type pat_aux =
    | PCon of user_id * pat list
    | PVar of user_id
    | PTuple of pat list
    | PWild
    | PZero
    | PAlt of pat * pat
  and pat = pat_aux * loc

  type expr_aux =
    | Var of user_id
    | Patch of patch_addr
    | Absurd
    | Let of is_meta * bool * (user_id * expr) list * expr
    | Match of expr * (pat * expr) list
    | Con of user_id * expr list
    | Abs of user_id * expr
    | App of expr * expr
    | Tuple of expr list
    | Label of string * expr
  and expr = expr_aux * loc
  and patch_addr =
    | DefaultPatch
    | PatchFor of ((user_id option * int option) * expr option) list
  and patch_ident =
    | PatchIdent of user_id
    | PatchSpecial of user_id
  and patch =
    | PatchHole of expr * patch_addr * expr
    | PatchLifting of patch_addr * patch_ident list
    | PatchOrnament of patch_addr * patch_ident list

  type tydef = TyDef of user_id * user_id list * (user_id * user_ty list) list

  type orn_def = {
    orn_name : user_id;
    orn_args : user_id list;
    orn_from : user_ty;
    orn_to : user_ty;
    orn_clauses : (pat * (user_id * user_ty) list * pat) list
  }


  type lift_def = {
    lift_from : user_id;
    lift_loc : loc;
    lift_to : user_id;
    lift_at : user_ty option;
    lift_with : (patch * loc) list;
  }

  type elift_def = {
    elift_from : id;
    elift_loc : loc;
    elift_to : id;
    elift_at : user_ty option;
    elift_with : (patch * loc) list;
  }

  let rec free_clause x (p,e) =
    not (bound_pat x p) && free_term x e
  and bound_pat x = function
      | PVar(x'), _ -> x=x'
      | PTuple(ps),_ | PCon(_,ps),_ -> List.exists (bound_pat x) ps
      | (PWild | PZero),_ -> false
      | PAlt(p1, p2), _ -> bound_pat x p1 || bound_pat x p2
  and free_term x = function
    | Var(x'), _ -> x = x'
    | Patch _, _ | Absurd, _ -> true
    | Let(_,false,bs,e),_  ->
       List.exists (fun (_,e) -> free_term x e) bs ||
         (List.for_all (fun (x',_) -> x <> x') bs && free_term x e)
    | Let(_,true,bs,e), _ ->
       List.for_all (fun (x',_) -> x <> x') bs &&
         (List.exists (fun (_,e) -> free_term x e) bs || free_term x e)
    | Match(e, cs), _ -> free_term x e || List.exists (free_clause x) cs
    | (Con(_, es) | Tuple(es)), _ -> List.exists (free_term x) es
    | Abs(x',e), _ -> x <> x' && free_term x e
    | App(e1,e2), _ -> free_term x e1 || free_term x e2
    | Label(_,e), _ -> free_term x e

  let rec visit (f: 'a -> pat -> 'a * pat) (init: 'a) (p: pat) : 'a * pat =
    let accum, p = f init p in
    let nf = visit f in
    let a, p' = match fst p with
      | PCon (a, l) ->
        let accum, l = List.fold_right (fun p (a, l) -> let a, p = nf a p in a, p :: l) l (accum, []) in
        accum, PCon (a, l)
      | PVar v -> accum, PVar v
      | PTuple l ->
        let accum, l = List.fold_right (fun p (a, l) -> let a, p = nf a p in a, p :: l) l (accum, []) in
        accum, PTuple l
      | PWild -> accum, PWild
      | PZero -> accum, PZero
      | PAlt (a, b) ->
        let accum, a = nf accum a in
        let accum, b = nf accum b in
        accum, PAlt(a, b)
    in
    a, (p', snd p)


    let rec visit_expr (f: 'b -> 'a -> expr -> 'b * 'a * expr) (init_depth: 'b) (init: 'a) (e: expr): 'a * expr =
      let init_depth, init, e = f init_depth init e in
      let e', b = e in
      let v = visit_expr f init_depth in
      let fold2 init l = List.fold_right (fun (a, c) (init, l) ->
        let init, c = v init c in
        init, ((a, c) :: l)) l (init, []) in
      let fold1 init l = List.fold_right (fun c (init, l) ->
        let init, c = v init c in
        init, (c :: l)) l (init, []) in
      let bu e' = e', b in
      match e' with
      | Absurd -> init, e
      | Var (_) -> init, e
      | Let (a, c, l, e) ->
        let init, l = fold2 init l in
        let init, e = v init e in
        init, Let(a, c, l, e) |> bu
      | Match (e, l) ->
        let init, e = v init e in
        let init, l = fold2 init l in
        init, Match(e, l) |> bu
      | Con (a, l) ->
        let init, l = fold1 init l in
        init, Con(a, l) |> bu
      | Abs (a, e) ->
        let init, e = v init e in
        init, Abs (a, e) |> bu
      | App (a, b) ->
        let init, a = v init a in
        let init, b = v init b in
        init, App(a, b) |> bu
      | Label (a, e) ->
        let init, e = v init e in
        init, Label(a, e) |> bu
      | Patch (PatchFor([l, Some e])) ->
        let init, e = v init e in
        init, Patch(PatchFor[l, Some e]) |> bu
      | Patch _ -> init, e
      | Tuple e ->
        let init, l = fold1 init e in
        init, Tuple l |> bu
end

module OrnImplicit = Implicit

include Implicit

type scheme = Scheme of id list * ty

type toplet = bool * (user_id  * expr) list

type stmt_aux =
  | StVal of toplet
  | StType of tydef list
  | StAbstractVal of user_id * user_ty (* Abstract values have a type, but cannot be ornamented (except identity ornament) *)
  | StOrnDef of orn_def list (* Mutual recursion ! *)
  | StLiftDef of lift_def list
  | StCommented of toplet
  | StPragma of user_id list
and stmt = stmt_aux * loc

let mk_tuple f = function
  | [(x,_)] -> x
  | l -> f l

module Explicit = struct
(* What about letrec/mu ? *)
  type pat' =
    | PCon of id * ty list * pat list
    | PVar of id
    | PWild
    | PAlt of pat * pat
    | PZero
  and pat = pat' * ty * loc


  let rec visit (f: 'a -> pat -> 'a * pat) (init: 'a) (p: pat) : 'a * pat =
    let accum,(p', b, c as p) = f init p in
    let nf = visit f in
    let a, p' = match p' with
      | PCon (a, tys, l) ->
        let accum, l = List.fold_right (fun p (a, l) -> let a, p = nf a p in a, p :: l) l (accum, []) in
        accum, PCon (a, tys, l)
      | PVar v -> accum, PVar v
      | PWild -> accum, PWild
      | PZero -> accum, PZero
      | PAlt (a, b) ->
        let accum, a = nf accum a in
        let accum, b = nf accum b in
        accum, PAlt(a, b)
    in
    a, (p', b, c)


  let _PTuple = function
    | [p,_,_] -> p
    | ps -> PCon(tuple_con.(List.length ps), List.map (fun (_,t,_) -> t) ps, ps)

  type mark = ..
  type mark += NiceName of string
  type mark += Unused of id list
  type mark += MatchClause of int * bool (* clause ordering and was a wildcard *)
  type mark += ConstrEqProtected

  type expr' =
    | Var of id * ty list
        (* all variables -- variables in scheme *)
    | Let of is_meta * bool * id list * (id * scheme * expr) list * expr
    | MultiFix of (id * (id * ty) * expr) list * id   (* Returns a tuple *)
    | Match of expr * (pat * expr) list
    | ShallowMatch of id * (id * id list * expr) list * expr (* the last expr corresponds to the absurd case *)
    | Con of id * ty list * expr list
    | ShallowCon of id * ty list * id list * id
    | Abs of id *  ty * expr
    | App of expr * expr
    | Mark of mark * expr
    | Absurd
    | Patch of patch_addr (* XXX this may be removed? *)
  and expr = expr' * ty * loc

  type 'a visitor_folding = All of 'a | Each of 'a list

  let rec visit_expr (f: 'b -> 'a -> expr -> 'b visitor_folding * 'a * expr) (init_depth: 'b) (init: 'a) (e: expr): 'a * expr =
    let init_depth, init, e = f init_depth init e in
    let e', a, b = e in
    let v i = match init_depth with All init_depth -> visit_expr f init_depth i | Each _ -> assert false in
    let v' = visit_expr f in
    let fold3 init l =
      let l = match init_depth with
      | All a -> List.map (fun i -> a, i) l
      | Each l' -> List.combine l' l
      in
      List.fold_right (fun (id, (a, b, c)) (init, l) ->
      let init, c = v' id init c in
      init, ((a, b, c) :: l)) l (init, []) in
    let fold2 init l =
      let l = match init_depth with
      | All a -> List.map (fun i -> a, i) l
      | Each l' -> List.combine l' l
      in
      List.fold_right (fun (id, (a, c)) (init, l) ->
      let init, c = v' id init c in
      init, ((a, c) :: l)) l (init, []) in
    let fold1 init l =
      let l = match init_depth with
      | All a -> List.map (fun i -> a, i) l
      | Each l' -> List.combine l' l
      in
      List.fold_right (fun (id, c) (init, l) ->
      let init, c = v' id init c in
      init, (c :: l)) l (init, []) in
    let bu e' = e', a, b in
    match e' with
    | Patch _ | Absurd | Var (_, _) -> init, e
    | Let (a, b, c, l, e) ->
      let init, l = fold3 init l in
      let init, e = v init e in
      init, Let(a, b, c, l, e) |> bu
    | MultiFix (a, b) ->
      let init, a = fold3 init a in
      init, MultiFix(a, b) |> bu
    | Match (e, l) ->
      let init, e = v init e in
      let init, l = fold2 init l in
      init, Match(e, l) |> bu
    | ShallowMatch (a, l, absurd_case) ->
      let init, absurd_case = v init absurd_case in
      let init, l = fold3 init l in
      init, ShallowMatch(a, l, absurd_case) |> bu
    | Con (a, b, l) ->
      let init, l = fold1 init l in
      init, Con(a, b, l) |> bu
    | ShallowCon (_, _, _, _) ->
      init, e
    | Abs (a, b, e) ->
      let init, e = v init e in
      init, Abs (a, b, e) |> bu
    | App (a, b) ->
      let init, a = v init a in
      let init, b = v init b in
      init, App(a, b) |> bu
    | Mark (m, a) ->
      let init, a = v init a in
      init, Mark (m, a) |> bu

    let rec visit_expr_out (f: 'b -> 'a -> expr -> 'b * 'a * expr) (init_depth: 'b) (init: 'a) (e: expr): 'a * expr =
      (*let init_depth, init, e = f init_depth init e in*)
      let e', a, b = e in
      let v = visit_expr_out f init_depth in
      let fold3 init l = List.fold_right (fun (a, b, c) (init, l) ->
        let init, c = v init c in
        init, ((a, b, c) :: l)) l (init, []) in
      let fold2 init l = List.fold_right (fun (a, c) (init, l) ->
        let init, c = v init c in
        init, ((a, c) :: l)) l (init, []) in
      let fold1 init l = List.fold_right (fun c (init, l) ->
        let init, c = v init c in
        init, (c :: l)) l (init, []) in
      let bu e' = e', a, b in
      let init, e = match e' with
      | Absurd | Patch _ | Var (_, _) -> init, e
      | Let (a, b, c, l, e) ->
        let init, l = fold3 init l in
        let init, e = v init e in
        init, Let(a, b, c, l, e) |> bu
      | MultiFix (a, b) ->
        let init, a = fold3 init a in
        init, MultiFix(a, b) |> bu
      | Match (e, l) ->
        let init, e = v init e in
        let init, l = fold2 init l in
        init, Match(e, l) |> bu
      | ShallowMatch (a, l, absurd_case) ->
        let init, absurd_case = v init absurd_case in
        let init, l = fold3 init l in
        init, ShallowMatch(a, l, absurd_case) |> bu
      | Con (a, b, l) ->
        let init, l = fold1 init l in
        init, Con(a, b, l) |> bu
      | ShallowCon (_, _, _, _) ->
        init, e
      | Abs (a, b, e) ->
        let init, e = v init e in
        init, Abs (a, b, e) |> bu
      | App (a, b) ->
        let init, a = v init a in
        let init, b = v init b in
        init, App(a, b) |> bu
      | Mark (m, a) ->
        let init, a = v init a in
        init, Mark (m, a) |> bu in
      let _, a, b = f init_depth init e in a, b

  let visit_expr_dfs (f: 'a -> expr -> 'a * expr) (i: 'a) (e: expr) : 'a * expr =
    visit_expr_out (fun () a b -> let a, b = f a b in (), a, b) () i e


  let _Tuple = function
    | [e,_,_] -> e
    | es -> Con(tuple_con.(List.length es), List.map (fun (_,t,_) -> t) es, es)

  let _LetTuple = function
    | ([y,t],x,e) -> Let(IsShare, false, [], [(y,Scheme([],t),(Var(x,[]),t,dummy_loc))], e)
    | (ys,x,e) -> ShallowMatch(x, [tuple_con.(List.length ys), List.map fst ys, e], e)

  type toplet = TopLet of bool * (id * scheme *  expr) list

  type tydef = TyDef of bool (* is_global *) * (id * id list * (id * ty list) list) list

  type stmt_aux =
    | StVal of toplet
    | StType of tydef
    | StAbstractVal of id * scheme
    | StCommented of toplet
    | StLiftDef of lift_def
    | StOrnDef of orn_def list
  and stmt = stmt_aux * loc

  let mark m ((_ ,ty,loc) as e) = (Mark(m,e), ty, loc)

  let patvars (t : pat) : id list =
    let merge l1 l2 =
      List.iter (fun x1 -> if List.mem x1 l2 then failwith "Duplicate variable in pattern matching") l1;
      l1@l2
    in
    let rec vars (p,_,_) = match p with
      | PCon(_,_,ps) -> List.fold_left merge [] (List.map vars ps)
      | PVar (x) -> [x]
      | PWild -> []
      | PZero -> []
      | PAlt(p,q) ->
         let vp, vq = vars p, vars q in
         let equal = List.sort compare vp = List.sort compare vq in
         if not equal
         then failwith "Different variables in both sides of an alternative pattern"
         else vp
    in
    vars t

type orn =
  | OArrow of orn * orn
  | OIdOrn of id * orn list
  | OBaseOrn of id * orn list
  | OVar of id
  | OProduct of orn list

let oty_free_vars =
  let rec ftv = function
    | OVar v -> IdSet.singleton v
    | OArrow(t1,t2) -> IdSet.union (ftv t1) (ftv t2)
    | OProduct(ts) | OIdOrn(_,ts) | OBaseOrn(_,ts) ->
       List.fold_left (fun s t -> IdSet.union s (ftv t)) IdSet.empty ts
  in
  ftv

(* The soup is like this so we can impose the identity ornament... *)
module Soup = struct
  (* orn_soup is the output of the inference *)
  type orn =
    | OId of id * orn list
    | OArrow of orn * orn
    | OProduct of orn list
    | OCon of id * orn list * orn
    | OVar of id
    | OCycle of id * orn
end

let soup_subst_var x y =
  let rec su = function
    | Soup.OVar(i) -> Soup.OVar(if x = i then y else i)
    | Soup.OArrow(w1, w2) -> Soup.OArrow(su w1, su w2)
    | Soup.OProduct(ws) -> Soup.OProduct(List.map su ws)
    | Soup.OCycle(i, w) -> Soup.OCycle(i, su w)
    | Soup.OCon(c, ws, w) -> Soup.OCon(c, List.map su ws, su w)
    | Soup.OId(c, ws) -> Soup.OId(c, List.map su ws)
  in
  su

let unsoup =
  let rec us s = function
    | Soup.OId _ -> Error.failwith "Bad structure"
    | Soup.OVar(i) -> OVar(i)
    | Soup.OArrow(w1, w2) -> OArrow(us s w1, us s w2)
    | Soup.OProduct(ws) -> OProduct(List.map (us s) ws)
    | Soup.OCycle(v,t) -> us (IdSet.add v s) t
    | Soup.OCon(_,_, inst) -> us2 s inst
  and us2 s = function
    | Soup.OCycle(v,t) -> us2 (IdSet.add v s) t
    | Soup.OId(c,ws) -> OIdOrn(c, List.map (us s) ws)
    | Soup.OVar(v) ->
       if IdSet.mem v s
       then Error.failwith "Bad cycle: was there a cyclic type in the original term?"
       else OVar(v)
    | _ -> Error.failwith "Bad structure"
  in
  us IdSet.empty
(*
let rec unsoup = function
  | Soup.OId _ -> Error.failwith "Bad structure"
  | Soup.OVar(i) -> OVar(i)
  | Soup.OArrow(w1, w2) -> OArrow(unsoup w1, unsoup w2)
  | Soup.OProduct(ws) -> OProduct(List.map unsoup ws)
  | Soup.OCycle(_,Soup.OCon(_,_,inst))
  | Soup.OCon(_,_, inst) ->
     begin match inst with
           | Soup.OId(c, ws) -> OIdOrn(c, List.map unsoup ws)
           | Soup.OVar(v) -> OVar(v)
           | _ -> Error.failwith "Bad structure"
     end
  | Soup.OCycle(_, Soup.OArrow(_)) -> Error.failwith "the bad cycle was an arrow"
  | Soup.OCycle _ -> Error.failwith "Bad cycle: was there a cyclic type in the original term?"
*)

type orn_def = {
  orn_name : id;
  orn_args : id list;
  orn_from : ty;
  orn_from_skel : scheme; (* Should be: (...) * orn. we should remove the ty/orn distinction. *)
  orn_to : ty;
  orn_clauses : (pat * pat) list;
  orn_fallback : bool
}

let get_all_orn_args o =
  let Scheme(l, _) = o.orn_from_skel in
  l @ o.orn_args

module ExplicitGml = struct
  type expr' =
    | Var of id
    | Absurd
    | Let of is_meta * (id * ty * expr) list * expr
    | MultiFix of (id * (id * ty) * expr) list * id
    | Match of expr * (id * id list * expr) list * expr
    | Con of id * ty list * expr list
    | Tuple of ty list * expr list
    | LetTuple of (id * ty) list * expr * expr
    | Abs of id * ty* expr
    | App of expr * expr
    | GAbs of id * expr
    | GApp of expr * expr (* Should be a pseudo-value *)
    | GAbsT of id * expr (* Kinds disappeared :) *)
    | GAppT of expr * ty
    | Label of string * expr
    | Mark of mark * expr
   and expr = expr' * ty * loc
   and ty =
    | TFall of id * ty
    | TGFall of id * ty
    | TMatch of expr * (id * id list * ty) list
    | TCon of id * ty list
    | TVar of id
    | TArr of ty * ty
    | TGArr of id * ty * ty
    | TAbs of id * ty * ty
    | TApp of ty * expr
    | TAbsT of id * ty
    | TAppT of ty * ty
    | TProd of ty list

  let unit_ty = TCon(id "unit", [])
  let unit_term = Con(id "()", [], []), unit_ty, dummy_loc

  let rec of_ml_ty = function
    | TyArrow(t1,t2) -> TArr(of_ml_ty t1, of_ml_ty t2)
    | TyCon(n, ts) -> TCon(n, List.map of_ml_ty ts)
    | TyVar(v) -> TVar(v)
    | TyProduct(ts) -> TProd(List.map of_ml_ty ts)

  let of_ml_scheme (Scheme(vs,ty)) =
    List.fold_right (fun v t -> TFall(v,t)) vs (of_ml_ty ty)

  (* Assume full subst, no collisions *)
  let close_ty subst =
    let rec cl = function
      | TVar(v) -> List.assoc v subst
      | TCon(c, ts) -> TCon(c, List.map cl ts)
      | TArr(t1,t2) -> TArr(cl t1, cl t2)
      | _ -> failwith "Not designed for this"
    in
    cl

end

module E = ExplicitGml

type instance_request =
  (* An onrnament variable:
    Source location, hole number, ornament variable, ornament of (iota omega* ), variables: ext, con, proj
  *)
  | InstantiateOrnVar of loc * int * id * id * orn list * id * id * id
  (* identity ornament of (iota omega* ), variables: ext, con, proj *)
  | IdentityOrnament of id * orn list * id * id * id
  (* variables in scope, location, hole id, variables for the patch, type *)
  | InstantiatePatch of id list * loc * int * id list * E.ty
  (* A function from the environment:
     location, hole id, variable, original variable,
     instance for polymorphic vars, instance for ornament vars *)
  | InstantiateEnv of loc * int * id * id * orn list * orn list

type ornscheme = {
  ornscheme_paravars : id list;
  ornscheme_ornvars : (id * id * orn list) list;
  ornscheme_ornty : orn
}

type skeleton_desc = {
  skel_args : id list;
  skel_id_args : id list;
  skel_id_inst : ty list;
  skel_conses : (id * ty list) list
}

type generic_env = {
  skel_map : (id * id) list; (* Associates a skeleton type/data constructor to a constructor *)
  skel_unmap : (id * id) list;
  skel_dataenv : (id * skeleton_desc) list;
  skel_conenv : (id * id) list; (* Gives the datatype from the constructor *)
  (* What does an ornamentation scheme look like ?
      * Variables that are completely free
      * Things that are ornamentation-of-something
   *)
  gen_varenv : (id * ornscheme) list
}

module Implicit = OrnImplicit

type lifting_descr = {
  lifting_of : id;
  lifting_targs : id list;
  lifting_insts : orn list;
  lifting_orns : orn list;
}

type generic_descr = {
  generic_tvars : id list;
  generic_ornscheme : ornscheme;
  generic_requests : instance_request list;
  generic_rec_renaming : (id * id) list;
  generic_term : E.expr;
}

type orn_env = {
  env_ornaments : (id * (orn_def * E.ty * E.expr * E.expr)) list;
  (* indexed by relation: important because they may be multiple relations with the same destination
     (id ornament !) *)
  env_liftings : ((id * id) * lifting_descr) list;
  orn_scope : Scope.scope;
  (* A generic ornament: the list of free type/ornament variables, the scheme, the instance requests, the generic term *)
  orn_generic : (id * generic_descr) list
}

let initial_orn_env = { env_ornaments = []; env_liftings = []; orn_scope = Scope.empty_scope; orn_generic = [] }


  module SubstRename = struct
    (* Substitutes free variables using the given substitution,
       rename all bound variables to use fresh names *)

    let make_subst_full subst_te subst_ty =
      let open List in
      let h = IdHashtbl.create 17 in
      let bind x = let x' = fresh_id x in
                   IdHashtbl.add h x x'; x' in
      let rn_te x tys = try Var(IdHashtbl.find h x,tys) with Not_found -> subst_te x tys in
      let rn_ty x = try TyVar(IdHashtbl.find h x) with Not_found -> subst_ty x in
      let rn_var x = try IdHashtbl.find h x with Not_found -> let Var(v,[]) = subst_te x [] in v in
      let rec rte_ = function
        | Var(x, tys) -> rn_te x (List.map rty tys)
        | Let(is_meta,is_rec,scvars,bindings,e) ->
           let scvars = map bind scvars in
           iter (fun (x,_,_) -> ignore (bind x)) bindings;
           let bindings = map (fun (x,Scheme(qvs,ty),e) -> (rn_var x, Scheme(List.map rn_var qvs, rty ty), rte e)) bindings in
           Let(is_meta,is_rec,scvars,bindings,rte e)
        | MultiFix(bindings, returned) ->
           List.iter (fun (x,_,_) -> ignore (bind x)) bindings;
           MultiFix(List.map (fun (x,(a,t),e) ->
                              let a = bind a in
                              (rn_var x, (a, rty t), rte e)
                             ) bindings,
                    rn_var returned)
        | Match(e,clauses) ->
           let clauses = map (fun (p,e) ->
                         List.iter (fun x -> ignore (bind x)) (patvars p);
                         (rtp p, rte e)
                        ) clauses in
           Match(rte e, clauses)
        | ShallowMatch(x, clauses, e) ->
           ShallowMatch(rn_var x, List.map (fun (c,xs,e) -> let xs = List.map bind xs in (c,xs,rte e)) clauses, rte e)
        | Con(c,ts,es) -> Con(c, map rty ts, map rte es)
        | ShallowCon(c,ts,xs,conid) -> ShallowCon(c, List.map rty ts, List.map rn_var xs,conid)
        | Abs(x,ty,e) -> let x = bind x in Abs(x, rty ty, rte e)
        | App(e1,e2) -> App(rte e1, rte e2)
        | Mark(Unused l, e) -> Mark (Unused (List.map rn_var l), rte e)
        | Mark(m,e) -> Mark(m, rte e)
        | Absurd -> Absurd
        | Patch p -> Patch p
      and rte (e,ty,loc) = (rte_ e, rty ty, loc)
      and rtp_ = function
        | PCon(c, ts, ps) -> PCon(c, List.map rty ts, List.map rtp ps)
        | PVar(x) -> PVar(rn_var x)
        | PWild -> PWild
        | PAlt(p,q) -> PAlt(rtp p, rtp q)
        | PZero -> PZero
      and rtp (p,ty,loc) = (rtp_ p, rty ty, loc)
      and rty  = function
        | TyArrow(t1,t2) -> TyArrow(rty t1, rty t2)
        | TyCon(c, ts) -> TyCon(c, List.map rty ts)
        | TyVar(x) -> rn_ty x
        | TyProduct(ts) -> TyProduct(List.map rty ts)
      in
      (rte, rty,rtp)
    let id_term x tys = Var(x,tys)
    let id_type x = TyVar x
    let make_subst te ty = let (rte,rty,_) = make_subst_full te ty in (rte, rty)
    let fresh_term te = let (rte,_,_) = make_subst_full id_term id_type in rte te
    let fresh_type ty = let (_,rty,_) = make_subst_full id_term id_type in rty ty
  end
end

type dataenv = (id * (id list * (id * ty list) list)) list
type conenv = (id * (id * id list * ty list)) list


type env = { conenv : conenv;  varenv : (id * scheme) list; dataenv : dataenv; scope : Scope.scope}

module Orn = Explicit

let untype (tl : Explicit.toplet) : toplet =
  let open Explicit in
  let rec term (t,_,loc) = (term_aux loc t, loc)
  and pat (p,_,loc) = (pat_aux p, loc)
  and term_aux loc : Explicit.expr' -> Implicit.expr_aux = function
    | Var(v, _) -> Var(FromId v)
    | Absurd -> Absurd
    | Patch p -> Patch p
    | Let(is_meta, isrec, _, bindings, k) ->
       Let(IsUser, isrec, List.map (fun (x,_,t) -> (FromId x, term t)) bindings, term k)
    | Match(e,cs) -> Match(term e, List.map (fun (p,e) -> pat p, term e) cs)
    | Con(i,_,ts) -> Con(FromId i, List.map term ts)
    | Abs(x,_,t) -> Abs(FromId x, term t)
    | App(u,v) -> App(term u, term v)
    | MultiFix(bindings, ret) ->
       Let(IsShare, true,
           List.map (fun (x,(y,_),t) -> (FromId x, (Implicit.Abs(FromId y, term t), loc))) bindings,
           (Var(FromId ret), loc)
          )
    | Mark(NiceName n, e) -> Label(n, term e)
    | Mark(_, e) -> fst (term e)
    | (ShallowMatch _  | ShallowCon _) -> assert false (* should have been resugared first *)
  and pat_aux : Explicit.pat' -> Implicit.pat_aux = function
    | PVar v -> PVar (FromId v)
    | PCon(c,_,ps) -> PCon(FromId c, List.map pat ps)
    | PWild -> PWild
    | PZero -> PZero
    | PAlt(p1, p2) -> PAlt(pat p1, pat p2)
  in
  match tl with
  | TopLet(is_rec, bs) -> (is_rec, List.map (fun (x,_,t) -> FromId x, term t) bs)

(* Call when eliminating an "let x = y in ..." *)
let register_subst x y =
  y.id_equals <- x::y.id_equals

let register_expr_subst x (e,_,_) =
  match e with
  | Explicit.Var(y,_) -> register_subst x y
  | _ -> ()
