open Ast.Explicit
type id = Ast.id

let var_eq_modulo_alpha l v1 v2 =
  v1.Id.id_name = "__" || v1 = v2 ||
  try
    List.assoc v1 l = v2
  with Not_found -> false

let rec eq_modulo_alpha_aux ?equalize:(equalize = (fun _ -> ())) (l: (id * id) list) (e1:expr) (e2:expr) =
  let e1', _, _ = e1 in
  let e2', _, _ = e2 in
  let eq_modulo_alpha_aux = eq_modulo_alpha_aux ~equalize in
  match e1', e2' with
  | Var(i1, _), Var(i2, _) -> var_eq_modulo_alpha l i1 i2
  | Let(_, _, _, bound_vars1, e1), Let(_, _, _, bound_vars2, e2) ->
    if List.length bound_vars1 <> List.length bound_vars2 then
      false
    else begin
      let bound_vars = List.combine bound_vars1 bound_vars2 in
      let new_l = List.map (fun ((i1, _, e1), (i2, _, e2)) ->
          i1, i2) bound_vars in
      List.iter equalize new_l;
      let l = new_l @ l in
      List.fold_left (fun is_eq ((_, _, e1), (_, _, e2)) ->
          is_eq && eq_modulo_alpha_aux l e1 e2) (eq_modulo_alpha_aux l e1 e2) bound_vars end
  | MultiFix (funs1, ret1), MultiFix (funs2, ret2) ->
    if List.length funs1 <> List.length funs2 then
      false
    else begin
      let funs = List.combine funs1 funs2 in
      let new_l = List.map (fun ((i1, _, e1), (i2, _, e2)) ->
          i1, i2) funs in
      let l = new_l @ l in
      List.iter equalize new_l;
      let new_l = List.map (fun ((_, (v1, _), e1), (_, (v2, _), e2)) ->
          v1, v2) funs in
      List.iter equalize new_l;
      let l = new_l @ l in
      List.fold_left (fun is_eq ((_, _, e1), (_, _, e2)) ->
          is_eq && eq_modulo_alpha_aux l e1 e2) (var_eq_modulo_alpha l ret1 ret2) funs end
  | Con (n1, _, args1), Con (n2, _, args2) ->
    begin try
        List.combine args1 args2
        |> List.fold_left (fun is_eq (e1, e2) ->
            is_eq && eq_modulo_alpha_aux l e1 e2) (n1 = n2)
      with
      | Invalid_argument _ -> false
    end
  | ShallowCon (c1, _, args1, _), ShallowCon (c2, _, args2, _) -> (* FIXME: not sure what the last argument of ShallowCon is supposed to be *)
    begin try
        List.combine args1 args2
        |> List.fold_left (fun is_eq (e1, e2) ->
            is_eq && var_eq_modulo_alpha l e1 e2) (c1 = c2)
      with
      | Invalid_argument _ -> false
    end
  | Abs (v1, _, e1), Abs(v2, _, e2) ->
    equalize (v1, v2);
    let l = (v1, v2) :: l in
    eq_modulo_alpha_aux l e1 e2
  | App(a1, b1), App(a2, b2) ->
    eq_modulo_alpha_aux l a1 a2 && eq_modulo_alpha_aux l b1 b2
  | Mark(_, e1), Mark(_, e2) ->
    eq_modulo_alpha_aux l e1 e2
  | Absurd, Absurd -> true
  | Patch _, Patch _ -> true
  | Match (e, cls), Match (e', cls') ->
    eq_modulo_alpha_aux l e e' && List.fold_left2 (fun is_eq (p1, e1) (p2, e2) ->
        is_eq && eq_pat_aux ~equalize l p1 p2 (fun l -> eq_modulo_alpha_aux l e1 e2)) true cls cls'
  | ShallowMatch (arg1, branches1, abs1), ShallowMatch (arg2, branches2, abs2) ->
    let sort_branch = List.sort (fun (a, _, _) (a', _, _) -> compare a a') in
    var_eq_modulo_alpha l arg1 arg2 &&
    List.fold_left (fun is_eq ((con1, args1, e1), (con2, args2, e2)) ->
        is_eq &&
        con1 = con2 &&
        (let new_l = List.combine args1 args2 in
         List.iter equalize new_l;
         let l = new_l @ l in
         eq_modulo_alpha_aux l e1 e2)
      ) true (List.combine (sort_branch branches1) (sort_branch branches2)) &&
    eq_modulo_alpha_aux l abs1 abs2
  | _ -> false
and eq_pat_aux ~equalize l p1 p2 f =
  match p1, p2 with
  | _, (PWild, _, _) -> f l
  | (PWild, _, _), _ -> f l
  | (PAlt(p1_, p1'), _, _), (PAlt(p2_, p2'), _, _) ->
    eq_pat_aux ~equalize l p1_ p2_ (fun l ->
        eq_pat_aux ~equalize l p1' p2' f)
  | (PCon(v, _, args), _, _), (PCon(v', _, args'), _, _) when v = v' ->
    (List.fold_left2 (fun f p1 p2 l ->
         eq_pat_aux ~equalize l p1 p2 f) f args args') l
  | (PVar v, _, _), (PVar v', _, _) ->
    equalize (v, v');
    f ((v, v')::l)
  | _ -> false


let eq_modulo_alpha (e1 : expr) (e2 : expr) : bool =
  eq_modulo_alpha_aux [] e1 e2

let extract_all_used_marks (e : expr) : Id.IdSet.t =
  let open Id in
  visit_expr_dfs (fun l e ->
      match e with
      | Mark(Unused l', _), _, _ -> IdSet.union (IdSet.of_list l') l, e
      | _ -> l, e) IdSet.empty e
  |> fst

let remove_unused_marks (e : expr) : expr =
  visit_expr_dfs (fun () e ->
      match e with
      | Mark(Unused _, e), _, _ -> (), e
      | _ -> (), e) () e
  |> snd

let rec dead_code_elim_no_fp ?mark:(mark=false) (e: expr) =
  let open Id in
  let used_vars = if mark then IdSet.empty else extract_all_used_marks e in
  visit_expr_dfs (fun (used_vars : Id.IdSet.t) e ->
      let e', ty, loc = e in
      let e = match e' with
        | Let (a, b, c, bound_vars, e1) ->
          if List.exists (fun (a, _, _) -> IdSet.mem a used_vars) bound_vars then
            e
          else if mark then
            let vars = List.map (fun (a, _, _) -> a) bound_vars in
            Let (a, b, c, bound_vars, (Mark(Unused vars, e1), ty, loc)), ty, loc
          else
            e1
        | MultiFix (funs, ret) ->
          if mark then
            let funs_unused =
              List.filter (fun (n, _, _) ->
                  not (IdSet.mem n used_vars || ret = n)) funs
              |> List.map (fun (n, _, _) -> n) in
            Mark(Unused funs_unused, e), ty, loc
          else
            let funs = List.filter (fun (n, _, _) ->
                IdSet.mem n used_vars || ret = n) funs in
            MultiFix(funs, ret), ty, loc
        | ShallowMatch (_, branches, abs) ->
          (* first, check that no variable is used *)
          let args_are_needed = List.exists (fun (conid, args, e) ->
              List.exists (fun a -> IdSet.mem a used_vars) args) branches in
          if not args_are_needed then
            (* now, check that all branches are equal to  the first one *)
            match branches with
            | [] -> Format.eprintf "empty match@."; assert false
            | (_, _, first)::q ->
              if List.exists (fun (_, _, e') ->
                  not (eq_modulo_alpha first e')) q then e
              else if mark then e else first
          else e
        | Match (_, _) -> assert false
        | _ -> e in
      let e', ty, loc = e in
      let used_vars = IdSet.union (match e' with
          | Var (i, _) -> IdSet.singleton i
          | ShallowMatch (a, _, _) -> IdSet.singleton a
          | ShallowCon (_, _, l, _) -> IdSet.of_list l
          | _ -> IdSet.empty) used_vars in
      used_vars, e
    ) used_vars e
  |> snd

let rec fixpoint (f: expr -> expr) (a: expr): expr =
  let b = f a in
  if b = a then a
  else fixpoint f b

let dead_code_elim ?mark:(mark=false) e =
  fixpoint (dead_code_elim_no_fp ~mark) e
