open Util
(* Identifiers and renaming *)

type id = {
  id_name : string;
  id_unique : int;
  id_copy : id option;
  id_user : bool;
  mutable id_equals : id list
}

(* The FromId ids must be globally unique, because they will be imported as-is *)
type user_id = UserId of string | FromId of id

let id =
  let c = ref 0 in
  fun s -> incr c; { id_name = s; id_unique = !c; id_copy = None; id_equals = []; id_user = false}

let good_id s = debug "good_id %s@." s; {(id s) with id_user = true}
let fresh_id i = {(id i.id_name) with id_copy = Some i; id_user = i.id_user}
let user_id s = UserId(s)

module TeVar = struct
  type tevar = id
  let compare i1 i2 =
(*    let c = compare i1.id_name i2.id_name in
    if c = 0 then (* assert (i1 = i2) fails ? *) compare i1.id_unique i2.id_unique
      else c *) compare i1.id_unique i2.id_unique
  let show i = i.id_name^"@"^string_of_int i.id_unique
end

module IdHash = struct
  type t = id
  let hash i = i.id_unique land max_int
  let equal i j = i.id_unique = j.id_unique
  let compare i j = compare i.id_unique j.id_unique
end

module IdSet = Set.Make(IdHash)
module IdMap = Map.Make(IdHash)
module IdHashtbl = Hashtbl.Make(IdHash)

module PrettyNames = struct
  (* Take a bunch of id, and convert them to user_ids *)
  (* Function: scope (fun  -> ...),
       pretty (id),
       global (id)
  *)
  (* Do a first pass to determine the id -> name mapping,
     then rename everything in a second pass *)
  (* ie first pass, just accumulate variable names and return dummies
     second pass do the actual attribution (because we know what names
     are forbidden for us) *)
  type ns = NsVar | NsCons | NsTy | NsTyVar
  type tag = Intro | IntroGlobal | Use

  type scope_p1 = {
    mutable vars : (id * tag) list;
    mutable conses : (id * tag) list;
    mutable types : (id * tag) list;
    mutable tyvars : (id * tag) list;
    mutable children : scope_p1 list;
  }
  type scope = S1 of scope_p1 | S2 of (id * user_id) list

  let scope s f =
    match s with
      | S1(s) ->
        let s' = {vars=[];conses=[];types=[];tyvars=[];children=[]} in
        s.children <- s'::s.children;
        f (S1 s')
      | S2(m) -> f (S2 m)

  let insert tag s ns x =
    match s with
      | S1(s) ->
        begin
          match ns with
            | NsVar -> s.vars <- (x,tag)::s.vars
            | NsTy -> s.types <- (x,tag)::s.types
            | NsCons -> s.conses <- (x,tag)::s.conses
            | NsTyVar -> s.tyvars <- (x,tag)::s.tyvars
        end;
        UserId("<dummy>")
      | S2(m) -> List.assoc x m

  let id = insert Use
  let intro_global = insert IntroGlobal
  let intro = insert Intro

  let find_best_name x =
    let rec find x =
      if x.id_user
      then Some(x.id_name)
      else first_some (bind_option find x.id_copy) (try_search x.id_equals find)
    in
    match find x with
    | Some(n) -> n
    | None -> (debug "no good name found for %s@." (TeVar.show x); x.id_name)

  (* Call "f", list free identifiers, then call f again in "repair" mode *)
  let renamer do_warn s =
    (* propagate mutates s to keep the list of alive variables,
       and return the list of alive variables minus introduced there too *)
    let merge_val x y = match x,y with
      | Use, x | x, Use -> x
      | Intro, Intro -> Intro
      | IntroGlobal, x | x, IntroGlobal -> IntroGlobal
    in
    let merge vs =
      let vs = List.sort compare vs in
      let rec ana = function
        | [] -> []
        | [x] -> [x]
        | (v,x)::(v',y)::xs when v = v' -> ana ((v,merge_val x y)::xs)
        | (v,x)::xs -> (v,x)::ana xs
      in
      ana vs
    in
    let filter vs = List.filter (fun (_,x) -> x = Use) vs in
    (* List the variables, merge the values *)
    let rec propagate s =
      let subs = List.map propagate s.children in
      let vars,types,conses,tyvars =
        List.concat (List.map (fun (x,_,_,_) -> x) subs),
        List.concat (List.map (fun (_,x,_,_) -> x) subs),
        List.concat (List.map (fun (_,_,x,_) -> x) subs),
        List.concat (List.map (fun (_,_,_,x) -> x) subs)
      in
      s.vars <- merge (s.vars @ vars);
      s.types <- merge (s.types @ types);
      s.conses <- merge (s.conses @ conses);
      s.tyvars <- merge (s.tyvars @ tyvars);
      (filter s.vars, filter s.types, filter s.conses, filter s.tyvars)
    in
    let compose x = function
      | 0 -> x
      | n -> x^string_of_int n
    in
    let rec assign (uvars,utypes,uconses,utyvars) s =
      (* Basically: keep a list of used names, use the first open variation on
         the name *)
      let treat used s =
        (* Remove all variables that can be shadowed *)
        let used = List.filter (fun (_,v) -> List.exists (fun (v',_) -> v=v') s) used in
        (* Assign variables *)
        List.fold_left (fun (used,map) ->
          function
            | _, Use -> (used,map)
            | ({id_name = stem; _} as v,IntroGlobal) ->
              if List.mem_assoc stem used
              then (Format.eprintf "Failed to satisfy renaming problem for global %s@." stem; failwith "renaming error")
              else ((stem,v)::used, (v,UserId(stem))::map)
            | (v, Intro) ->
               let stem = find_best_name v in 
               let rec name i =
                 let n = compose stem i in
                 if List.mem_assoc n used
                 then name (i+1)
                 else n
               in
               let n = name 0 in
               ((n,v)::used, (v,UserId(n))::map)
        ) (used,[]) s
      in
      let (uvars, mvars) = treat uvars s.vars in
      let (utypes, mtypes) = treat utypes s.types in
      let (uconses, mconses) = treat uconses s.conses in
      let (utyvars, mtyvars) = treat utyvars s.tyvars in
      mvars @ mtypes @ mconses @ mtyvars @ List.concat (List.map (assign (uvars,utypes,uconses,utyvars)) s.children)
    in
    let (u,v,w,x) = propagate s in
    let id ppf i = Format.fprintf ppf "%s@@%d" i.id_name i.id_unique in
    List.iter (fun (v,_) -> Format.eprintf "Leftover: %a@." id v) (u@v@w@x);
    if not (u = [] && v = [] && w = [] && x = []) && do_warn
    then Format.eprintf "Warning: unbound variables (probably because the lifting is incomplete)@.";
    let f x = List.map (fun (x,_) -> (x, UserId x.id_name)) x in
    assign ([],[],[],[])  s @ f u @ f v @ f w @ f x

  let run_renamer ?warn:(do_warn=true) f =
    let s = {vars=[];tyvars=[];conses=[];types=[];children=[]} in
    ignore (f (S1 s));
    let m = renamer do_warn s in
    f (S2 m)

end

let sum_ty = good_id "either"
let sum_l = good_id "Left"
let sum_r = good_id "Right"

let max_tuple_size = 16
let tuple_ty = Array.init max_tuple_size (fun i -> if i = 0 then good_id "unit" else good_id ("tuple"^string_of_int i))
let tuple_con = Array.init max_tuple_size (fun i -> if i = 0 then good_id "Unit" else good_id ("Tuple"^string_of_int i))

let tuple_names =
  let h = IdHashtbl.create 129 in
  for i = 0 to max_tuple_size - 1 do
    IdHashtbl.add h tuple_ty.(i) ();
    IdHashtbl.add h tuple_con.(i) ();
  done;
  h

let is_tuple_name u =
  IdHashtbl.mem tuple_names u
let check_tuple n = if n >= max_tuple_size then failwith "Only tuples with length < 16 are supported"
