{
  open Lexing
  open Parser

  let incr_loc lexbuf delta =
    let pos = lexbuf.Lexing.lex_curr_p in
    lexbuf.Lexing.lex_curr_p <- { pos with
      Lexing.pos_lnum = pos.Lexing.pos_lnum + 1;
      Lexing.pos_bol = pos.Lexing.pos_cnum - delta;
    }

  let update_loc lexbuf opt_file line =
    let pos = lexbuf.Lexing.lex_curr_p in
    let new_file = match opt_file with
                   | None -> pos.Lexing.pos_fname
                   | Some f -> f
    in
    lexbuf.Lexing.lex_curr_p <- { pos with
      Lexing.pos_fname = new_file;
      Lexing.pos_lnum = line;
      Lexing.pos_bol = pos.Lexing.pos_cnum;
    }
  let is_con s = String.length s > 0 && 'A' <= s.[0] && 'Z' >= s.[0]

  let id_or_kwd = function
    | "match" -> Match
    | "with" -> With
    | "let" -> Let
    | "fun" -> Fun
    | "rec" -> Rec
    | "and" -> And
    | "in" -> In
    | "type" -> Type
    | "of" -> Of
    | "ornament" -> Ornament
    | "relation" -> Ornament
    | "lifting" -> Lifting
    | "patch" -> Patch
    | "begin" -> Begin
    | "end" -> End
    | "val" -> Val
    | "when" -> When
    | "function" -> Function
    | s -> if is_con s then Con s else Ident s

  exception Lexing_error of string
}

let id = ['a'-'z' 'A'-'Z'] ['\'' 'a' - 'z' 'A' - 'Z' '_' '0' - '9']* | '_' ['\'' 'a' - 'z' 'A' - 'Z' '_' '0' - '9']+
let space = [' ' '\t']+
let number = ['0'-'9']+

rule main = parse
  | "\n" { Lexing.new_line lexbuf; main lexbuf }
  | id as i { id_or_kwd i }
  | "(" { Lparen }
  | ")" { Rparen }
  | "[" { Lbrack}
  | "]" { Rbrack }
  | "->" { Arrow }
  | "=>" { DoubleArrow }
  | "<-" { RevArrow }
  | "=" { Equal }
  | "|" { Pipe }
  | "'" { Quote }
  | "," { Comma }
  | "*" { Star }
  | ":" { Colon }
  | ";" { Semicolon }
  | "_" { Underscore }
  | "~" { Tilde }
  | "#" { Sharp }
  | "@" { At }
  | number as n { IntLit(int_of_string n) }
  | space { main lexbuf }
  | "#" [' ' '\t']+ (['0'-'9']+ as num) [' ' '\t']*
    ('\"' ([^ '\010' '\013' '\"']* as name) '\"')?
    [^ '\010' '\013']* '\010'
    { update_loc lexbuf name (int_of_string num);
      main lexbuf
    }
  | eof { EOF }
  | "(*" { comment 0 lexbuf }
and comment i = parse
  | '\n' { Lexing.new_line lexbuf; comment i lexbuf }
  | "(*" { comment (i+1) lexbuf }
  | "*)" { if i = 0 then main lexbuf else comment (i-1) lexbuf }
  | _ { comment i lexbuf }
  | eof { raise (Lexing_error "Unterminated comment") }

{

 let lex_with_errors file lexbuf =
    try let r = main lexbuf in r
    with Lexing_error e -> Error.(error (Loc.lex_loc file lexbuf) "Lexing error: %s" e)
       | e -> Error.(error (Loc.lex_loc file lexbuf) "Error during lexing: %s" (Printexc.to_string e))

}
