open UnifierSig

module S = struct

  type 'a structure =
    | TyArrow of 'a * 'a
    | TyCon of Ast.id * 'a list


  let map f t =
    match t with
    | TyArrow (t1, t2) ->
      let t1 = f t1 in
      let t2 = f t2 in
      TyArrow (t1, t2)
    | TyCon(s,l) -> TyCon(s,List.map f l)


  let fold f t accu =
    match t with
    | TyArrow (t1, t2) ->
      let accu = f t1 accu in
      let accu = f t2 accu in
      accu
    | TyCon (_,l) ->
      List.fold_left (fun x y -> f y x) accu l

  let iter f t =
    let _ = map f t in
    ()

  exception Iter2

  let iter2 f t u =
    match t, u with
    | TyArrow (t1, t2), TyArrow (u1, u2) ->
        f t1 u1;
        f t2 u2
    | TyCon(s1,l1), TyCon(s2,l2) when s1 = s2 ->
       if List.length l1 = List.length l2
       then List.iter2 f l1 l2
       else raise Iter2
    | TyCon(s1, _), TyCon(s2, _) ->
       raise Iter2
    | _, _ ->
        raise Iter2

  let show sub ppf = function
    | TyArrow(t1,t2) -> Format.fprintf ppf "%a -> %a" sub t1 sub t2
    | TyCon(c,p) -> Format.fprintf ppf "%a(%a)" Pp.id c (Pp.seq ", " sub) p
end

module O = struct

  type tyvar =
    Id.id

  type 'a structure =
      'a S.structure

  type ty =
    Ast.ty

  let variable x =
    Ast.TyVar x

  let mu x t = Ast.(TyVar(Id.id ("<cyclic>")))

  let structure = function
    | S.TyArrow(t1,t2) -> Ast.TyArrow(t1,t2)
    | S.TyCon(s,ts) -> Ast.TyCon(s,ts)

  type scheme =
      tyvar list * ty
end



module Solver = SolverHi.Make(Id.TeVar)(S)(O)

open Solver
open Ast
module X = Ast.Explicit

module UserIdSet = Set.Make(struct type t = user_id let compare = compare end)


let _ = Printexc.register_printer (function Unbound x -> Some("Solver: Unbound variable "^x.id_name^"@"^string_of_int x.id_unique) | _ -> None)

let conj_list l =
  List.fold_right (fun x y -> x ^& y >>= fun (x,y) -> x::y) l (pure [])


(* Create a mapping from type variables to bla variables *)
let freshen vars k =
  let rec remap acc k = function
    | [] -> k (List.rev acc) >>= fun x -> ([],x)
    | i::is -> exist (fun v -> remap ((i,v)::acc) k is) >>= fun (x,(xs,y)) -> (x::xs,y)
  in
  remap [] k vars


open Scope

(* returns a unit constraint *)
let rec ty_encode mapping tau : ty -> unit co = function
  | TyVar v ->
    tau --
      (try List.assoc v mapping
       with Not_found -> Format.eprintf "%a@." Pp.id v; failwith "unbound type variable")
  | TyArrow(t1,t2) ->
    exist_ (fun v1 -> exist_ (fun v2 ->
        ty_encode mapping v1 t1 ^^
          ty_encode mapping v2 t2 ^^
          tau --- S.TyArrow(v1,v2)
    ))
  | TyProduct _ -> assert false
  | TyCon(c,l) ->
    let rec check acc = function
      | [] -> tau --- S.TyCon (c, List.rev acc)
      | t::ts -> exist_ (fun v -> ty_encode mapping v t ^^ check (v::acc) ts)
                 >>= fun _ -> ()
    in
    check [] l

let encode_scheme vs ty v =
(* Add all exists ? *)
  let rec build acc = function
    | [] -> ty_encode acc v ty
    | x::xs -> exist_ (fun v -> build ((x,v)::acc) xs)
  in
  build [] vs

let patvars (t : Ast.pat) : user_id list =
  let merge l1 l2 =
    List.iter (fun x1 -> if List.mem x1 l2 then failwith "Duplicate variable in pattern matching") l1;
    l1@l2
  in
  let rec vars p = match fst p with
    | PTuple ts | PCon(_,ts) -> List.fold_left merge [] (List.map vars ts)
    | PVar (x) -> [x]
    | PWild -> []
    | PZero -> []
    | PAlt(p,q) ->
       let vp, vq = vars p, vars q in
       let equal = List.sort compare vp = List.sort compare vq in
       if not equal
       then failwith "Different variables in both sides of an alternative pattern"
       else vp
  in
  vars t

let not_wild = fun s ->
  match s with
  | UserId s ->
    let n = String.length s in
    not (n >= String.length "__wild" && String.sub s 0 (String.length "__wild") = "__wild")
  | _ -> assert false

let rec patvars_set (p: Ast.pat): UserIdSet.t =
  let open UserIdSet in
  match fst p with
  | PTuple ts | PCon(_, ts) ->
  List.map patvars_set ts
  |> List.fold_left (fun s1 s2 ->
      let n1 = cardinal s1 in
      let n2 = cardinal s2 in
      let s = union s1 s2 in
      if n1 + n2 <> cardinal s then failwith "Duplicate variable in pattern matching"
      else s) empty
  | PVar(x) -> singleton x
  | PWild | PZero -> empty
  | PAlt(p, q) ->
    let s1 = patvars_set p
    and s2 = patvars_set q in
    if exists not_wild (diff s1 s2) then
      failwith "Different variables in both sides of an alternative pattern"
    else union s1 s2

let is_value =
  let rec is_value (e, loc) = is_value' e
  and is_value' = function
    | Var _ -> true
    | Let(_,_, bindings, e) ->
       is_value e && List.for_all (fun (x, e) -> is_value e) bindings
    | Match(e, clauses) ->
       is_value e && List.for_all (fun (p, e) -> is_value e) clauses
    | Con(_, es) -> List.for_all is_value es
    | Abs(x, e) -> true
    | App(e1, e2) -> false
    | Tuple(es) -> List.for_all is_value es
    | Label(_,e) -> is_value e
    | Absurd | Patch _ -> true
  in
  is_value

let let_nonpolymorphic vars constr ret =
  (* Introduce variables for all in var, ... *)
  freshen vars
          (fun varmap ->
            constr (List.map (fun v -> List.assoc v varmap) vars) ^&
              List.fold_right (fun (x,v) -> def x v) varmap ret
          )
  >>= fun (ts, (a,b)) -> ([], a, List.map (fun t -> ([],t)) ts, b)

let inference_things e =
  let con scope hastype c ts tau =
    let c = resolve scope NsCons c in
    let (tname, tvars, args) = List.assoc c e.conenv in
    freshen tvars (fun mapping ->
      let ts_start = ts in
      let args_start = args in
      let rec quant args ts = match args, ts with
        | x::args, t::ts -> exist_ (fun v -> ty_encode mapping v x ^^ hastype t v ^& quant args ts) >>= fun (x,xs) -> x::xs
        | [], [] -> tau --- S.TyCon(tname,List.map snd mapping) >>= fun _ -> []
        | _ -> Format.eprintf "When looking at %a, lengths %d %d@." Pp.id c (List.length ts_start) (List.length args_start);
          failwith "arg length mismatch"
      in
      quant args ts
    ) >>= fun (vs,ts) -> (c,vs,ts)
  in

  let tuple (type tin) (type tout)
      (hastype : tin -> variable -> tout co) (ts : tin list) (tau : variable)
      : ('a list * tout list) co =
    check_tuple (List.length ts);
    let rec check acc = function
      | [] -> tau --- S.TyCon(tuple_ty.(List.length ts), List.rev acc)
              >>= fun _ -> ([],[])
      | t::ts -> exist (fun v -> hastype t v ^& check (v::acc) ts)
                 >>= fun (ty,(x,(tys,xs))) -> (ty::tys,x::xs)
    in
    check [] ts
  in
  let rec hastype scope t tau = exist (fun tau' -> tau -- tau' ^^ hastype' scope (fst t) tau ) >>= fun (ty,e) -> (e,ty,snd t)
  and hastype' scope (t : Ast.expr_aux) (tau : variable) : X.expr' co =
    match t with
    | Patch p ->
      pure (X.Patch p)
    | Absurd ->
      pure X.Absurd
    | Var x ->
      (* Find x in the env *)
      let x = resolve scope NsVar x in
      instance x tau >>= fun taus -> X.Var(x, taus)
    | App(t1, t2) ->
      exist (fun v ->
        lift (hastype scope) t1 (S.TyArrow(v,tau)) ^&
          hastype scope t2 v
      ) >>= fun (tau,(t1',t2')) ->
      X.App(t1', t2')
    | Abs(x,u) ->
      let (x,scope) = intro scope NsVar x in
      exist (fun v1 ->
        exist (fun v2 ->
          tau --- (S.TyArrow(v1,v2)) ^&
            def x v1 (hastype scope u v2)
        )
      ) >>= fun (tau1, (tau2, ((), u'))) ->
      X.Abs(x,tau1,u')
    | Let(is_meta, isrec,bindings,u) ->
      let is_polymorphic = List.for_all (fun (x,e) -> is_value e) bindings in
      let (bindings, newscope) = intros scope NsVar fst (fun (_,y) x -> (x,y)) bindings in
      let letscope = if isrec then newscope else scope in
      (if is_polymorphic then letn else let_nonpolymorphic) (List.map (fun (x,_) -> x) bindings)
        (fun vars ->
          let inner = conj_list (List.map2 (fun v (x,c) -> hastype letscope c v >>= fun c' -> (x,c')) vars bindings) in
          if isrec
          then List.fold_right2 (fun v (x,_) -> def x v) vars bindings inner
          else inner
        )
        (hastype newscope u tau)
      >>= fun (a,t',schemes,u') ->
        X.Let(is_meta,isrec,
              a,
              List.map2 (fun (x,c') (vs,ty) -> (x,Scheme(vs,ty), c'))
                t' schemes,
              u')
    | Match(e,cs) ->
      exist_ (fun vin ->
        hastype scope e vin ^&
          conj_list (List.map (fun t -> clause scope t vin tau) cs)
      ) >>= fun (e',cs') -> X.Match(e',cs')
    | Tuple ts -> tuple (hastype scope) ts tau >>= fun (tys,ts) -> X.Con(tuple_con.(List.length ts), tys, ts)
    | Con(c,ts) -> con scope (hastype scope) c ts tau >>= fun (c,vs,ts) -> X.Con(c,vs,ts)
    | Label(_,e) -> hastype scope e tau >>= fun (e,_,_) -> e
  and pattype scope t tau = exist (fun tau' -> tau -- tau' ^^ pattype' scope (fst t) tau) >>= fun (ty,p) -> (p,ty,snd t)
  and pattype' scope (t : Ast.pat_aux) (tau : variable) : X.pat' co =
    match t with
      | PTuple ts -> tuple (pattype scope) ts tau >>= fun (tys,ts) -> X.PCon(tuple_con.(List.length ts), tys, ts)
      | PCon(c,ts) -> con scope (pattype scope) c ts tau >>= fun (c,vs,ts) -> X.PCon(c,vs,ts)
      | PVar x ->
        let x = resolve scope NsVar x in
        instance x tau >>= fun ts -> assert (ts = []); X.PVar x
      | PWild -> pure X.PWild
      | PZero -> pure X.PZero
      | PAlt(p,q) -> (pattype scope p tau ^& pattype scope q tau) >>= fun (p,q) -> X.PAlt(p,q)
  and clause scope (t : Ast.pat * Ast.expr) (tau_in : variable) (tau_out : variable) :
      (X.pat * X.expr) co =
    let (p,e) = t in
    let xs = patvars p in
    let (xs,scope) = intros scope NsVar (fun x -> x) (fun _ x -> x) xs in
    let rec abstract k = function
      | [] -> k
      | x::xs -> exist_ (fun v -> def x v (abstract k xs))
    in
    abstract (pattype scope p tau_in ^& hastype scope e tau_out) xs
  in
  object
    method hastype = hastype
    method pattype = pattype
    method clause = clause
    method tuple = tuple
    method con = con
  end

let infer scope env isrec bindings =
  try
    let rec fillenv k = function
      | [] -> k
      | (x,Scheme(vs,ty))::xs ->
        let1 x
          (fun v -> encode_scheme vs ty v)
          (fillenv k xs)
        >>= fun (tvs,_,(tvs2,ty),r) -> r
    in
    (* Work on mutually recursive things ! *)
    solve false (let0 (fillenv
      (let is_polymorphic = List.for_all (fun (x,e) -> is_value e) bindings in
       if not is_polymorphic then Error.user_error "Non-polymorphic code at toplevel";
       let (bindings, newscope) = intros scope NsVar fst (fun (_,y) x -> (x,y)) bindings in
       let letscope = if isrec then newscope else scope in
       (if is_polymorphic then letn else let_nonpolymorphic) (List.map (fun (x,_) -> x) bindings)
         (fun vars ->
           let inner = conj_list (List.map2 (fun v (x,c) -> (inference_things env)#hastype letscope c v >>= fun c' -> (x,c')) vars bindings) in
           if isrec
           then List.fold_right2 (fun v (x,_) -> def x v) vars bindings inner
           else inner
         )
         (pure newscope)) env.varenv)
      >>= fun (vs, (_, t', schemes,newscope))  ->
List.map2 (fun (x,c) (vs,ty) -> (x, Scheme(vs,ty), c)) t' schemes, newscope
    )

  (* Intro a bunch of variables *)


                         (*
    solve false (let0 (fillenv (exist (fun v -> (inference_things env)#hastype scope t v)) env.varenv) >>=
                           fun (vs,(t,e)) -> ((vs,t),e) ) *)
  with
  | Unify(t1,t2) ->
     Error.user_error "Type error: cannot unify %a and %a" Pp.ty t1 Pp.ty t2


let resolve_ty scope =
  let rec res = function
    | TyVar(v) -> TyVar(resolve scope NsTyVar v)
    | TyCon(i,ts) -> TyCon(resolve scope NsType i, List.map res ts)
    | TyProduct(ts) ->
       let n = List.length ts in
       check_tuple n;
       TyCon(tuple_ty.(n), List.map res ts)
    | TyArrow(t1,t2) -> TyArrow(res t1, res t2)
  in
  res


let elaborate_tydefs scope defs =
  let defs, tyscope = intros scope NsType
                    (fun (TyDef(tn,_,_)) -> tn)
                    (fun (TyDef(_,args,conses)) tn -> (tn, args, conses)) defs in
  let defs, scope =
    List.fold_right (fun (tn,args,conses) (tl, scope) ->
                     let args, localscope = intros tyscope NsTyVar (fun x -> x) (fun _ x -> x) args in
                     let conses, scope = intros scope NsCons
                                                (fun (x,_) -> x)
                                                (fun (_,tys) x -> (x, List.map (resolve_ty localscope) tys))
                                                conses
                     in
                     ((tn, args, conses)::tl, scope))
                    defs ([], tyscope) in
  X.TyDef(true,defs), scope


let rec list_vars = function
  | TyVar(v) -> UserIdSet.singleton v
  | (TyCon(_,ts) | TyProduct(ts)) -> List.fold_left (fun s t -> UserIdSet.union s (list_vars t)) UserIdSet.empty ts
  | TyArrow(t1,t2) -> UserIdSet.union (list_vars t1) (list_vars t2)

let elaborate_abstract_def env x t =
  let x,scope = intro env.scope NsVar x in
  let vars = list_vars t in
  let (args, localscope) = intros scope NsTyVar (fun x -> x) (fun _ x -> x) (UserIdSet.elements vars) in
  let t = resolve_ty localscope t in
  let scheme = Scheme(args, t) in
  {env with scope = scope; varenv = (x, scheme)::env.varenv }, x, scheme

let make_recursive x e =
  Let(IsShare, true, [(x,e)], (Var(x), dummy_loc)), dummy_loc

exception Wrong_arity of loc

(*
    + constructors with a tuple as argument become n-ary constructors
 *)
let fix_ep_contuple env =
  let rec fix (e,loc) = (fix_aux loc e,loc)
  and fix_aux loc = function
    | Var _ | Patch _ | Absurd as e -> e
    | Let(is_meta,b, bs, e) -> Let(is_meta,b, List.map (fun (x,e) -> (x, fix e)) bs, fix e)
    | Match(e,cs) -> Match(fix e, List.map (fun (p,e) -> (pfix p, fix e)) cs)
    | Abs(x,e) -> Abs(x, fix e)
    | App(e1,e2) -> App(fix e1, fix e2)
    | Tuple(es) ->
       Tuple(List.map fix es)
    | Con(c,es) ->
       let c' = Scope.(resolve env.scope NsCons c) in
       let (_, _, ar) = List.assoc c' env.conenv in
       begin match ar, es with
             | [], [] -> Con(c,[])
             | _, [] -> raise (Wrong_arity loc)
             | [], _ -> raise (Wrong_arity loc)
             | [_], [e] -> Con(c, [fix e])
             | ar, [Tuple es, _] ->
                if List.length ar = List.length es
                then Con(c,List.map fix es)
                else raise (Wrong_arity loc)
             | _::_::_, [_] -> raise (Wrong_arity loc)
             | _, _::_::_ -> assert false (* Parser should never create this *)
       end
    | Label(l, e) -> Label(l, fix e)
  and pfix (p,loc) = (pfix_aux loc p,loc)
  and pfix_aux loc = function
    | (PVar _ | PWild | PZero) as p -> p
    | PTuple ps -> PTuple (List.map pfix ps)
    | PAlt(p1,p2) -> PAlt(pfix p1, pfix p2)
    | PCon(c, ps) ->
       let c' = Scope.(resolve env.scope NsCons c) in
       let (_, _, ar) = List.assoc c' env.conenv in
       begin match ar, ps with
             | [], [] -> PCon(c,[])
             | _, [] -> raise (Wrong_arity loc)
             | [], _ -> raise (Wrong_arity loc)
             | ar, [(PWild, _) as w] -> (* Cons _  -> Cons (_,_) *)
                PCon(c, List.map (fun _ -> w) ar)
             | [_], [p] -> PCon(c,[pfix p])
             | ar, [PTuple ps, _] ->
                if List.length ar = List.length ps
                then PCon(c,List.map pfix ps)
                else raise (Wrong_arity loc)
             | _::_::_, [_] -> raise (Wrong_arity loc)
             | _, _::_::_ -> assert false (* Parser should never create this *)
       end
  in
  (fix, pfix)

let fix_contuple env = fst (fix_ep_contuple env)
let pfix_contuple env = snd (fix_ep_contuple env)


let elaborate_def ?(extra_var = []) env isrec bindings =
  let scope = List.fold_left (fun scope id ->
    let x, scope = Scope.(intro scope NsVar (FromId id)) in
    assert (x = id); scope) env.scope extra_var in
  let env = { env with varenv = List.map (fun i -> let id = id "dummy" in i, Scheme([id], TyVar(id))) extra_var @ env.varenv; scope} in
(*  let (bindings, scope) = Scope.(intros env.scope NsVar (fun (x,_) -> x) (fun (_,e) x -> (x,e)) bindings) in
    let (x,scope) = Scope.(intro env.scope NsVar x) in *)
  let bindings, scope  = infer env.scope env isrec bindings in
  List.iter (fun (x,sc,e) ->
    say "typechecked [val %a : %a]@." Pp.id x Pp.tyscheme sc;
    debug " = %a@." Pp.xterm e
  ) bindings;
  {env with
    scope = scope;
    varenv = (List.map (fun (x,sc,_) -> (x,sc)) bindings)@env.varenv
  }, X.TopLet(isrec, bindings)
