(* From an ornament definition, generate the encoding of the ornament.
   Uses deep pattern matching elimination (\ie. in ML) before
   lifting to \gML in order to share the code of the elimination *)

open Ast
module O = Ast.Orn
open Util
module X = Ast.Explicit
module E = Eml.ExplicitGml
module I = Orn.Implicit
module S = Ast.Scope

let ty = TyArrow(TyVar (id "untyped"), TyVar (id "untyped"))
let sc = Scheme([],ty)

(* register this as an hole ! *)
let absurd_expr () =
  let x = id "_absurd_" in
  IdHashtbl.add Pretty.holemap x "<absurd>";
  X.Absurd, ty, dummy_loc

let rec pat_to_expr ?wild_to_unit:(wild_to_unit=false) = function
  | (X.PVar(x), ty, loc) -> X.Var(x,[]),ty,loc
  | (X.PCon(c,ts,ps), ty, loc) -> X.Con(c,ts, List.map (pat_to_expr ~wild_to_unit) ps), ty, loc
  | (X.PWild, _, loc) when wild_to_unit -> X.Con(Id.tuple_con.(0), [], []), ty, loc
  | _ -> failwith "Not a pattern and an expression"

(* Idea: somewhat flatten the sum-of-product for a more intuitive representation.
   Representation:
     sum([t]) -> t
     sum([unit, unit]) -> bool
     sum([unit,t])/sum([t,unit]) -> t option (?)
   otherwise, a either-telescope

  also, remove all units and impossible cases (if any ?)

  -- we should probably use polymorphic variants instead
 *)
type ext =
  | None
  | Unit
  | Thing


let opt_list = function
  | (None: 'a option) -> []
  | Some a -> [a]

let rec push_back f ?(e=(None: 'a option)) = function
  | [] -> opt_list e
  | e :: q when f e ->
    let e = Some e in
    push_back f ~e q
  | a :: q ->
    a :: push_back f ~e q

let push_back_left_wild = push_back (function
  | ((Explicit.PWild, _, _), _) -> true | _ -> false)

let push_back_right_wild = push_back (function
  | (_, ( (Explicit.PWild, _, _))) -> true | _ -> false)

let gen_code loc env def =
  (* Substitute on the left ! Compute the skeleton of the left side !
     Substitute on the right ! Keep the same type !
     Generate pattern matchings ! Simplify them !
     Substitute to get the functions !

     We are working with quasi-untyped code.
   *)
  let input = id "input" in
  let cvars = List.map (fun _ -> id "clause") def.O.orn_clauses in
  let ret_ty = TyVar(id "ret_ty") in
  (* Accidentally quadratic *)
  let rec prescan = function
    | X.PZero, _, _ -> None
    | X.PWild, _, _ -> Thing
    | X.PVar _, _ , _ -> Unit
    | X.PAlt(p1,p2), _, _ ->
       begin match prescan p1, prescan p2 with
       | None, t | t, None -> t
       | _ -> Thing
       end
    | X.PCon(_,_,ps), _, _ ->
       let cs = List.map prescan ps in
       if List.mem None cs then None
       else if List.for_all (fun x -> x = Unit) cs then Unit
       else Thing
  in
  (* Invariant: unit extensions do not look at the variable *)
  let rec extension (y,y_ty) = function
    | X.PWild, ty, loc -> X.Var(y,[]), ty, loc
    | X.PVar v, ty, loc -> X.Var(v,[]), ty, loc
    | X.PCon(c, ts, ps), ty, loc -> product_extension y ps ty loc (fun ls -> X.Con(c, ts, ls), ty, loc)
    | X.PZero, ty, loc -> absurd_expr ()
    | X.PAlt(p1, p2), ty, loc -> (* Might be nicer if we create n-way alternatives and n-way products *)
       begin match prescan p1, prescan p2 with
             | t, None -> extension (y, ty) p1
             | None, t -> extension (y, ty) p2
             | _ ->
                let y1, y2 = id "yl", id "yr" in
                X.Match( (X.Var(y,[]), ty, loc),
                         [(X.PCon(sum_l, [], [X.PVar(y1),ty,loc]),ty,loc), extension (y1, ty) p1;
                          (X.PCon(sum_r, [], [X.PVar(y2),ty,loc]),ty,loc), extension (y2, ty) p2] ), ty, loc
       end
  and product_extension y ps ty loc cont =
    let cs = List.map prescan ps in
    let ys = List.mapi (fun i (_,ty,_) -> id ("y"^string_of_int i), ty) ps in
    let u = cont (List.map2 extension ys ps) in
    let ys_truncated, _ = List.split (List.filter (fun (_,x) -> x <> Unit) (List.combine ys cs)) in
    if ys_truncated = [] then u
    else X._LetTuple(ys_truncated,y, u), ty, loc (* let (ys,) = y in u *)
  in
  let ext_var = id "ext_var" in
  let mkp p = (p, ty, dummy_loc) in
  let ty_from_skel = match def.O.orn_from_skel with Scheme(_, a) -> a in
  let left_side =
    X.Match((X.Var(input, []), ty_from_skel, dummy_loc),
            List.map2 (fun (p,q) v -> (p, extension (ext_var,ty) q)) (push_back_left_wild def.O.orn_clauses) cvars
              @ (if def.X.orn_fallback then [(mkp X.PWild, absurd_expr())] else [])
    ), ret_ty, dummy_loc in
  let right_side =
      X.Match((X.Var(input, []), def.O.orn_to, dummy_loc),
              List.map2 (fun (p,q) v -> match q with
                                        | X.PZero,_,_ -> (q, absurd_expr ())
                                        | q -> (q, extension (ext_var, ty) p)) (push_back_right_wild def.O.orn_clauses) cvars), ret_ty, dummy_loc in
  debug "Generated an ornament %a:\nLeft:\n%a\n\nRight:\n%a\n\n" Pp.id def.O.orn_name Pp.term left_side Pp.term right_side;
  let nice_encoder = left_side in
  let left_side =
    Desugaring.(
      try
        if true || def.X.orn_fallback then untyped_all env left_side
        else untyped_all_ornament env left_side
    with
    | PatternSimplification.OverlappingPatterns -> Error.error loc "Overlapping patterns on left-hand side of ornament definition"
    | PatternSimplification.IncompletePatterns -> Error.error loc "Incomplete pattern matching on left-hand side of ornament definition")
  in
  let right_side = Desugaring.(
    try
      untyped_all env right_side
    with
    | PatternSimplification.OverlappingPatterns -> Error.error loc "Overlapping patterns on right-hand side of ornament definition"
    | PatternSimplification.IncompletePatterns -> Error.error loc "Incomplete pattern matching on right-hand side of ornament definition")
  in
  (* Let us pretty-print the code before... *)
  say "Generated ornament %a with encoder:\n%a\n\n" Pp.id def.O.orn_name Pp.term nice_encoder;
  let open Eml_compute.ExplicitToGml in
  let orn_loc = dummy_loc in
  let dummy_ty = E.TVar(id "##UNTYPED##") in
  let left_side = E.GAbs(input, (E.GAbs(ext_var, ml_to_gml left_side), dummy_ty, orn_loc)), dummy_ty, orn_loc in
  let right_side = E.GAbs(input, (E.GAbs(ext_var, ml_to_gml right_side), dummy_ty, orn_loc)), dummy_ty, orn_loc in
  (*let right_side = E.GAbs(input, ml_to_gml right_side), dummy_ty, orn_loc in*)
  let ext_fun = E.TAbs(input, E.TVar(id "__input"), E.TVar(id "__ext")) in
  (* We theoretically want to return code in mML. So let us do this *)
  (ext_fun, left_side, right_side)

let check_left_pat lpat =
  let rec check (p,loc) =
    match p with
    | PVar _ -> ()
    | PWild -> ()
    | PAlt (a, b) -> check a; check b
    | PZero -> Error.error loc "Left pattern is not an expression (zero appears)"
    | PTuple ps -> List.iter check ps
    | PCon(_, ps) -> List.iter check ps
  in
  match lpat with
  | (PCon(_) | PTuple _ | PAlt _), _ -> check lpat
  | PZero, _ -> ()
  | _, loc -> Error.error loc "Left patterns must start with a constructor"

(* Take an ornament given in type-syntax, give its left projection *)
let oty_proj_left oenv loc =
  let rec proj = function
    | TyArrow(t1,t2) -> TyArrow(proj t1, proj t2)
    | TyVar(v) -> TyVar(v)
    | TyProduct(ts) -> TyProduct(List.map proj ts)
    | TyCon(c,ts) ->
       let odef,_,_,_ = List.assoc c oenv.X.env_ornaments in
       if List.length odef.X.orn_args <> List.length ts
       then Error.error loc "Wrong arity for %a." Pp.id c;
       let subst = List.map2 (fun x t -> (x,proj t)) odef.X.orn_args ts in
       let _, subst_ty = X.SubstRename.(make_subst id_term (fun v -> List.assoc v subst)) in
       subst_ty odef.X.orn_from
  in
  proj

(* Take an ornament given in type-syntax, give its right projection *)
let oty_proj_right oenv loc =
    let rec proj = function
    | TyArrow(t1,t2) -> TyArrow(proj t1, proj t2)
    | TyVar(v) -> TyVar(v)
    | TyProduct(ts) -> TyProduct(List.map proj ts)
    | TyCon(c,ts) ->
       let odef,_,_,_ = List.assoc c oenv.X.env_ornaments in
       if List.length odef.X.orn_args <> List.length ts
       then Error.error loc "Wrong arity for %a." Pp.id c;
       let subst = List.map2 (fun x t -> (x,proj t)) odef.X.orn_args ts in
       let _, subst_ty = X.SubstRename.(make_subst id_term (fun v -> List.assoc v subst)) in
       subst_ty odef.X.orn_to
  in
  proj

(* If a mapping is rigid, variables can only be mapped to variables *)
let invert_rigid mapping =
  let inv_map = List.map (function
    | x, TyVar y -> (y, x)
    | x, t -> Error.user_error "Type error: cannot unify type %a with type argument %a." Pp.ty_nice t Pp.id_nice x
  ) mapping in
  (* Check if the mapping is unique *)
  List.iter (fun (y,_) ->
    match List.filter (fun (y',_) -> y = y') inv_map with
    | (_,x)::(_,x')::_ -> Error.user_error "Type error: cannot unify type arguments %a and %a." Pp.id_nice x Pp.id_nice x'
    | _ -> ()
  ) inv_map;
  inv_map

let rec do_subst l (ty: ty) =
match ty with
| Ast.TyArrow (a, b) -> TyArrow (do_subst l a, do_subst l b)
| Ast.TyCon (c, l') -> TyCon(c, List.map (do_subst l) l')
| Ast.TyVar v -> (try List.assoc v l with Not_found -> TyVar v)
| Ast.TyProduct l' -> TyProduct (List.map (do_subst l) l')

let rec find_less_flexible ?(h=false) orn_args genv (ty: ty) =
  let open Ml_inference in
  let open Ml_inference.Solver in
  match ty with
  | TyArrow (a, b) ->
    TyArrow (find_less_flexible orn_args genv a, find_less_flexible orn_args genv b)
  | TyCon (con_skel, l) ->
    let l = List.map (find_less_flexible orn_args genv) l in
    begin try
      if h then raise Not_found;
      let skel_descr = List.assoc con_skel genv.X.skel_dataenv in
      let con = List.assoc con_skel genv.X.skel_unmap in
      let cstr =

        freshen orn_args (fun mapping' ->
        freshen skel_descr.X.skel_id_args (fun mapping ->
        construct (S.TyCon(con, List.map snd mapping)) (fun _ ->
          let mapping = mapping @ mapping' in
          freshen skel_descr.X.skel_args (fun skel_args ->
            let id_inst = List.combine skel_descr.X.skel_id_inst l in
            List.fold_left2 (fun a i (i2, i3) ->
              a ^^ ty_encode mapping i i2 ^^ ty_encode mapping i i3
              ) (Ml_inference.conj_list [] >>= fun _ -> ()) (List.map snd skel_args) id_inst
          )))) >>= fun (vtys, (_, (a, _))) ->
          let right_unmap = try invert_rigid (List.combine orn_args vtys) with _ -> raise Not_found in
          let _, unsubst_ty, unsubst_pa = X.SubstRename.(make_subst_full id_term (fun v ->
             TyVar(List.assoc v right_unmap))) in
          vtys, unsubst_ty a
      in
      solve false (let0 cstr) |> snd |> snd
    with
    | Unify _ -> TyCon(con_skel, l)
    | Not_found -> TyCon(con_skel, l) end
  | TyVar _ -> ty
  | TyProduct l -> assert false
    (*TyProduct (List.map find_less_flexible l)*)


let typecheck_orndef loc env genv oenv o =
  let orn_name = S.(resolve oenv.X.orn_scope NsType o.orn_name) in
  let orn_args, odef_scope = S.(intros oenv.X.orn_scope NsTyVar (fun x -> x) (fun _ x -> x) o.I.orn_args) in
  let orn_from = Ml_inference.resolve_ty odef_scope o.I.orn_from in
  let orn_to = Ml_inference.resolve_ty odef_scope o.I.orn_to in
  let i_skel = match orn_from with
    | TyCon(i, ts) -> List.assoc i genv.X.skel_map
    | _ -> Error.user_error "Not an ornament of a datatype"
  in
  let skel_descr = List.assoc i_skel genv.X.skel_dataenv in
  (* Check structure of left patterns *)
  let open Ml_inference in
  let open Ml_inference.Solver in
  let clauses = List.map (fun (lpat, otypes, rpat) ->
    (pfix_contuple env lpat, otypes, pfix_contuple env rpat)) o.orn_clauses in
  let orn_fallback, clauses =
    match List.rev clauses with
    | ((PWild,_),[],(PZero,_))::cs -> true, List.rev cs
    | _ -> false, clauses
  in
  List.iter (fun (lpat,otypes,rpat) ->
    let (_,loc) = lpat in
    check_left_pat lpat;
    begin match rpat with
    | PZero, _ -> ()
    | _ , loc ->
       (* Rule: no duplicate variables, except for variables starting with an underscore *)
       if
         UserIdSet.exists not_wild (UserIdSet.diff (patvars_set lpat) (patvars_set rpat))
       then Error.error loc "Variables do no match on left and right of pattern."
    end;
    let lvars = UserIdSet.union (patvars_set rpat) (patvars_set lpat) in
    List.iter (fun (x,_) -> if not (UserIdSet.mem x lvars)
      then Error.error loc "Variable %a is given a type but not bound by this pattern." Pp.user_id x
    ) otypes
  ) clauses;
  (* Typecheck left/right in the new environment *)
  let cstr_to_along_ty ty_target =
    freshen orn_args (fun mapping ->
      let rec abstract k = function
        | [] -> k
        | (x,oty)::xs ->
          exist_ (fun v -> ty_encode mapping v oty ^^ def x v (abstract k xs))
      in
      let rec abstract_no_type k = function
        | [] -> k
        | (x,x')::xs -> exist_ (fun v -> def x' v (abstract_no_type k xs))
      in
      freshen skel_descr.X.skel_args (fun l ->
        construct (S.TyCon(i_skel, List.map snd l)) (fun skel_v ->
          exist_ (fun from_v ->
            ty_encode mapping from_v orn_from ^^
        exist_ (fun to_v ->
          ty_encode mapping to_v orn_to ^^
            conj_list
            (List.map
               (fun (lpat, otypes, rpat) ->
                 let lpat_skel =
                 let is_skel = function
                   | Some (TyCon(c, _)) ->
                    begin try List.assoc c genv.X.skel_unmap |> ignore; true
                    with Not_found -> false end
                    | _ -> true
                 in
                 let extract_skel_args ps = function
                   | Some (TyCon(c, l)) ->
                     let ty_args = List.assoc c env.dataenv |> fst in
                     let l = List.combine ty_args l in
                     List.map (fun (id, ty) ->
                      (match id with Some id ->
                        let t = (do_subst l id) in
                        Some t
                        | None -> None), ty) ps
                   | Some(TyProduct l) -> assert false (*
                      List.map2 (fun (_, ty) l ->
                      Format.eprintf "tuple %a@." Pp.ty l;
                      Some l, ty) ps l*)
                   | None -> List.map (fun (id, a) -> (None:'a option), a) ps
                   | Some t -> Format.eprintf "%a@." Pp.ty t; assert false
                 in

                  let rec aux_compute_skel ty = function
                  | PCon(c, ps), loc ->
                    let c = Scope.(resolve env.scope NsCons c) in
                    let c_skel = if is_skel ty then List.assoc c genv.X.skel_map else c in
                    let _, _, cons_args = List.assoc c_skel env.conenv in
                    let cons_args = List.map (fun s -> Some s) cons_args in
                    let args = extract_skel_args (List.combine cons_args ps) ty in
                    let ps = List.map (fun (ty, a) -> aux_compute_skel ty a) args in
                    PCon(FromId c_skel, ps), loc
                  | PTuple(ps), loc ->
                    let c = tuple_con.(List.length ps) in
                    let c_skel = if is_skel ty then List.assoc c genv.X.skel_map else c in
                    let cons_args = if is_skel ty then let _, _, a = List.assoc c_skel env.conenv in a else
                    List.assoc tuple_ty.(List.length ps) env.dataenv|> snd |> List.hd |> snd in
                    let cons_args = List.map (fun s -> Some s) cons_args in
                    let args = extract_skel_args (List.combine cons_args ps) ty in
                    let ps = List.map (fun (ty, a) -> aux_compute_skel ty a) args in
                    PCon(FromId c_skel, ps), loc
                  | PAlt(a, b), loc ->
                    PAlt(aux_compute_skel ty a, aux_compute_skel ty b), loc
                  | a -> a
                   in
                   aux_compute_skel ty_target lpat
                 in
                 let all_vars =
                   patvars_set lpat
                   |> UserIdSet.union (patvars_set rpat)
                   |> UserIdSet.elements
                   |> List.map (fun x ->
                   match List.assoc x otypes with
                   | oty ->
                      let oty = resolve_ty odef_scope oty in
                      x, Some oty
                   | exception Not_found -> x, None)
                 in
                 let vars_ornamented, vars_same =
                   List.partition (function | _, Some _ -> true | _ -> false) all_vars
                  in
                 let vars_ornamented = List.map (function | x, Some a -> x, a | _ -> assert false) vars_ornamented in

                 let otypes, pat_scope = Scope.(intros env.scope NsVar (fun (x,oty) -> x) (fun (_,oty) x' ->
                  (x',oty)) vars_ornamented) in
                 let vars_same, pat_scope = Scope.(intros pat_scope NsVar (fun x -> x) (fun x x' -> x, x') (List.map fst vars_same)) in
                 let right_types = List.map (fun (x,oty) -> (x, oty_proj_right oenv loc oty)) otypes in
                 let left_types = List.map (fun (x,oty) -> (x, oty_proj_left oenv loc oty)) otypes in
                 abstract_no_type
                 (abstract ((inference_things env)#pattype pat_scope lpat from_v) left_types ^&
                 abstract ((inference_things env)#pattype pat_scope lpat_skel skel_v) otypes ^&
                 abstract  ((inference_things env)#pattype pat_scope rpat to_v) right_types) vars_same
               ) clauses)
        )
      )
    ))) >>= fun (vtys, (_, (skel_from, clauses))) ->
    let variables_left = ref [] in
    let right_unmap = invert_rigid (List.combine orn_args vtys) in
    let _, unsubst_ty, unsubst_pa = X.SubstRename.(make_subst_full id_term (fun v ->
      try TyVar(List.assoc v right_unmap)
      with Not_found ->
        variables_left := v :: !variables_left; TyVar v)) in
    let skel_from = unsubst_ty skel_from in
    let clauses =
    List.map (fun (_, (lpat_skel, rpat)) ->
      let lpat_skel = unsubst_pa lpat_skel in
      let rpat = unsubst_pa rpat in
      lpat_skel, rpat) clauses in
      skel_from, clauses, !variables_left
  in
  let _, (orn_from_skel, orn_clauses, variables) =
    try
      solve false (let0 (cstr_to_along_ty None))
    with
    | Unify(t1,t2) -> Error.error loc "Type error in ornamentation clause: cannot unify %a and %a" Pp.ty_nice t1 Pp.ty_nice t2
  in
  let ty_flexible = find_less_flexible ~h:true (orn_args @ variables) genv orn_from_skel in
  let _, (orn_from_skel, orn_clauses, variables) =
    try
      solve false (let0 (cstr_to_along_ty (Some ty_flexible)))
    with
    | Unify(t1,t2) -> Error.error loc "Type error in ornamentation clause: cannot unify %a and %a (should *not* happen)" Pp.ty_nice t1 Pp.ty_nice t2
  in
  say "[type ornament _ %a : %a => %a] is defined@." Pp.id_nice orn_name Pp.ty_nice orn_from_skel Pp.ty_nice orn_to;
  let variables = List.sort_uniq compare variables in
  let orn_from_skel = Scheme (variables, orn_from_skel) in
  let orn_def = X.{orn_name; orn_args;orn_from;orn_from_skel;orn_to;orn_clauses;orn_fallback} in
  orn_def

let rm_suffix suf s =
  let n = String.length suf
  and n' = String.length s in
  if n' < n then (None: string option)
  else if String.sub s 0 n = suf then Some (String.sub s n (n' - n) )
  else None

let reverse_user_id (odef_scope, orns) a =
  match a with
  | UserId s ->
    let orn_name = S.(resolve odef_scope NsType a) in
    if List.mem orn_name orns then
      let a' = UserId ("rev_" ^ s) in
      (* Now we check that the ornament was defined, it could be identity *)
      try S.(resolve odef_scope NsType a') |> ignore; a'
      with _ ->
        match rm_suffix "rev_" s with
        | Some s -> UserId s
        | _ -> (* then this is an identity ornament and it inverse is itself *) a
    else a
  | a -> assert false

let rec reverse_user_ty orns (v: Ast.user_ty) = match v with
  | TyArrow (a, b) -> TyArrow(reverse_user_ty orns a, reverse_user_ty orns b)
  | TyCon (s, a) -> TyCon (reverse_user_id orns s, List.map (reverse_user_ty orns) a)
  | TyVar u -> TyVar(u)
  | TyProduct a -> TyProduct  (List.map (reverse_user_ty orns) a)

let elaborate_orndef loc env genv oenv os =
  (* Generate a typed version of the `os': code, and skeleton arguments *)
  (* Introduce ornament names without the code at first *)
  (* Make the reverse ornament. We disable it if we spot an empty pattern (whose reverse has not been investigated yet) *)
  let make_rev = List.fold_left (fun make_rev o ->
      List.fold_left (fun make_rev (_, _, rpat) ->
        make_rev && (match rpat with PVar _, _ -> false | _ -> true)
      ) make_rev o.orn_clauses
    ) true os in
  let every_os =
    (if make_rev then List.map (fun o ->
      match o.I.orn_name with
      | UserId s -> UserId ("rev_" ^ s)
      | _ -> assert false) os
      else []) @ List.map (fun o -> o.I.orn_name) os in
  let _, odef_scope = S.(intros oenv.Orn.orn_scope NsType (fun o -> o) (fun _ _ -> ())) every_os in
  let all_orns = (oenv.Orn.env_ornaments |> List.map fst) @ (List.map (fun o ->
    S.(resolve odef_scope NsType o)
    ) every_os) in
  let os = List.fold_right (fun os l ->
    os :: begin if make_rev then
      let orn_name = match os.orn_name with
      | Ast.UserId s -> UserId ("rev_" ^ s)
      | Ast.FromId _ -> assert false in
      let orn_clauses = List.map (fun (lpat, vars, rpat) ->
        let vars = List.map (fun (id, v) ->
          id, reverse_user_ty (odef_scope, all_orns) v
          ) vars in
        rpat, vars, lpat) os.orn_clauses in
      let rev_os = { orn_name; orn_from = os.orn_to; orn_to = os.orn_from; orn_clauses; orn_args = os.orn_args; } in
      rev_os :: l
      else l end) os [] in
  let orns_stub = List.map (fun o ->
    let orn_name = S.(resolve odef_scope NsType o.orn_name) in
    let orn_args, odef_scope = S.(intros odef_scope NsTyVar (fun x -> x) (fun _ x -> x) o.I.orn_args) in
    let orn_from = Ml_inference.resolve_ty odef_scope o.I.orn_from in
    let orn_to = Ml_inference.resolve_ty odef_scope o.I.orn_to in
    let orn_from_skel = Scheme([], TyVar(id "dummy")) in
    let orn_clauses = [] in
    let orn_fallback = false in
    let dummy_ty = E.TVar(id "dummy") in
    let dummy_expr = E.Var(id "dummy"), dummy_ty, dummy_loc in
    orn_name,
    (X.{ orn_name; orn_args; orn_from; orn_from_skel; orn_to; orn_clauses; orn_fallback },
     dummy_ty, dummy_expr, dummy_expr)
  ) os in
  let oenv_stub = Orn.{oenv with orn_scope = odef_scope; env_ornaments = orns_stub@oenv.env_ornaments} in
  let os = List.map (typecheck_orndef loc env genv oenv_stub) os in
  (* Typecheck in the stub environment, generate code and add to the normal environment *)
  let orns = List.map (fun o -> let (ext,left,right) = gen_code loc env o in
                                o.X.orn_name, (o, ext, left, right) ) os in
  Orn.{oenv with orn_scope = odef_scope; env_ornaments = orns@oenv.env_ornaments}

let orndef_identity env generic_env i =
  (* I only use generic_env *)
  let i_skel = try List.assoc i generic_env.X.skel_map with Not_found -> i in
  let skel_desc = List.assoc i_skel generic_env.X.skel_dataenv in
  let orn_args = if i_skel == i then skel_desc.X.skel_args
                 else skel_desc.X.skel_id_args in
  let ty_args = List.map (fun x -> TyVar(x)) orn_args in
  let orn_ty = TyCon(i, ty_args) in
  let skel_args = if i == i_skel then ty_args else skel_desc.X.skel_id_inst in
  let orn_ty_skel = TyCon(i_skel, skel_args) in
  let clauses =
    List.map
      (fun (c_skel, args_ty) ->
       let c = if i_skel == i then c_skel else List.assoc c_skel generic_env.X.skel_unmap in
       let vars = List.map (fun ty -> (X.PVar(id "x"), TyVar (id "dummy"), dummy_loc)) args_ty in
       (X.PCon(c_skel, skel_args, vars), orn_ty_skel, dummy_loc),
       (X.PCon(c, ty_args, vars), orn_ty, dummy_loc)
      ) skel_desc.X.skel_conses
  in
  let orn_from_skel = Scheme([], orn_ty_skel) in
  X.{
    orn_name = i;
    orn_args;
    orn_from = orn_ty;
    orn_from_skel;
    orn_to = orn_ty;
    orn_clauses = clauses;
    orn_fallback = false
  }

let elaborate_identity loc env genv oenv is =
  let orn_scope =
    List.fold_left (fun orn_scope i -> Scope.(intro_named orn_scope NsType (UserId i.id_name) i))
      oenv.X.orn_scope is in
  let orns = List.map (fun i ->
    let idef = orndef_identity env genv i in
    let (ext,left,right) = gen_code loc env idef in
    (idef.X.orn_name, (idef, ext, left, right))
  ) is in
  Orn.{oenv with env_ornaments = orns@oenv.env_ornaments; orn_scope}

(* Ideally we would want to type the patch but this is simpler *)
let name_patch scope vars e =
  (* Filter out non-user variables. They should not be used in patches *)
  debug "Vars where %a@." Pp.(seq "," id) vars;
  let vars = List.filter (fun x -> x.id_user || (String.length x.id_name > 0 && x.id_name.[0] = '_')) vars in
  debug "Filtered to %a@." Pp.(seq "," id) vars;
  let open Scope in
  let scope = {scope with vars = List.map (fun x -> (UserId(x.id_name), x)) vars @ scope.vars} in
  let ty = TyVar(id "") in
  let sc = Scheme([], ty) in
  let rec cv_aux scope = function
    | Var(x) -> X.Var(resolve scope NsVar x, [])
    | Let(is_meta,isrec, bs, e) ->
       let vars, in_scope = intros scope NsVar (fun (x,_) -> x) (fun _ x -> x) bs in
       let let_scope = if isrec then in_scope else scope in
       X.Let(is_meta, isrec, [], List.map2 (fun x (_,e) -> (x, sc, cv let_scope e)) vars bs, cv in_scope e)
    | Match(e, cs) ->
       X.Match(cv scope e, List.map (clause scope) cs)
    | Con(c,es) -> X.Con(resolve scope NsCons c,[], List.map (cv scope) es)
    | Abs(x, e) ->
       let x, scope = intro scope NsVar x in
       X.Abs(x, ty, cv scope e)
    | App(e, f) -> X.App(cv scope e, cv scope f)
    | Tuple(es) -> X._Tuple(List.map (cv scope) es)
    | Label(l, e) -> X.Mark(X.NiceName l, cv scope e)
    | Patch p -> X.Patch p
    | Absurd -> X.Absurd
  and cv scope (e,loc) = cv_aux scope e, ty, loc
  and clause scope (p, e) =
    let vars = Ml_inference.patvars p in
    let _, scope = intros scope NsVar (fun x -> x) (fun _ x -> x) vars in
    (pat scope p, cv scope e)
  and pat scope(p, loc) = pat_aux scope p, ty, loc
  and pat_aux scope = function
    | PZero -> X.PZero
    | PVar(x) -> X.PVar(resolve scope NsVar x)
    | PWild -> X.PWild
    | PAlt(p,q) -> X.PAlt(pat scope p, pat scope q)
    | PTuple(ps) -> X._PTuple(List.map (pat scope) ps)
    | PCon(c,ps) -> X.PCon(resolve scope NsCons c, [], List.map (pat scope) ps)
  in
  cv scope e
