(* Ornament inference
   Principles: no concrete ornaments, only abstract
   (we keep a mapping between variables and kinds of onrnaments somewhere)
   Returns the ornamentation environment and the generic lifting
 *)

(* We use a particular structure:
   In the unification, an ornamnent phi of (o1, ...on) t is
   represented by phi <| (o1, ...on) t,
   \ie. the head of the structure is t.
   The phi is instanciated either by a variable (that will be an ornament variable)
   or by the constant "id".
   We cheat (how ?) and keep an association between phi and (o1, ...on) t in memory
   because it is necessary for rebuilding the ornamentation environment from a bunch
   of variables.

   We identify the ornament-types with the lifted-types. That's ok for identity
   and for ornament variables, which is all we use.
 *)

open UnifierSig
open Ast.Orn
module Orn = Ast.Orn

module S = struct

  (* XXX this is not the ideal representation of our structure... XXX *)
  type 'a structure =
    | TyArrow of 'a * 'a
    | TyProduct of 'a list
    | TyCon of Ast.id * 'a list * 'a
    | TyId of Ast.id * 'a list


  let map f t =
    match t with
    | TyArrow (t1, t2) ->
      let t1 = f t1 in
      let t2 = f t2 in
      TyArrow (t1, t2)
    | TyProduct (l) -> TyProduct(List.map f l)
    | TyCon(s,l,t) -> TyCon(s,List.map f l, f t)
    | TyId(s, l) -> TyId(s, List.map f l)


  let fold f t accu =
    match t with
    | TyArrow (t1, t2) ->
      let accu = f t1 accu in
      let accu = f t2 accu in
      accu
    | TyProduct (l) | TyId(_, l) -> List.fold_left (fun x y -> f y x) accu l
    | TyCon(_,l,t) -> f t (List.fold_left (fun x y -> f y x) accu l)

  let iter f t =
    let _ = map f t in
    ()

  exception Iter2

  let iter2 f t u =
    match t, u with
    | TyArrow (t1, t2), TyArrow (u1, u2) ->
        f t1 u1;
        f t2 u2
    | TyProduct(l1), TyProduct(l2) ->
      (try List.iter2 f l1 l2
      with _ when List.length l1 <> List.length l2 -> raise Iter2)
    | TyCon(s1,l1,t1), TyCon(s2,l2,t2) when s1 = s2 ->
      (try List.iter2 f l1 l2; f t1 t2
       with _ when List.length l1 <> List.length l2 -> raise Iter2)
    | TyId(s1, l1), TyId(s2,l2) when s1 = s2 ->
       (try List.iter2 f l1 l2
       with _ when List.length l1 <> List.length l2 -> raise Iter2)
    | _, _ ->
        raise Iter2

  let show sub ppf = function
    | TyArrow(t1,t2) -> Format.fprintf ppf "%a -> %a" sub t1 sub t2
    | TyProduct(p) -> Format.fprintf ppf "(%a)" (Pp.seq ", " sub) p
    | TyCon(c,p,t) -> Format.fprintf ppf "(%a(%a) >> %a)" Pp.id c (Pp.seq ", " sub) p sub t
    | TyId(s, l) -> Format.fprintf ppf "(%a(%a))" Pp.id s (Pp.seq ", " sub) l
end

module O = struct

  type tyvar =
    Id.id

  type 'a structure =
      'a S.structure

  type ty =
    Orn.Soup.orn

  let variable x =
    Orn.Soup.OVar x

  let mu v t =
    match v with
    | Orn.Soup.OVar(i) -> Orn.Soup.OCycle(i,t)
    | _ -> assert false (* I don't understand the interface here. Why not directly a variable ? *)

  let structure = function
    | S.TyArrow(t1,t2) -> Orn.Soup.OArrow(t1,t2)
    | S.TyCon(s,ts,t) -> Orn.Soup.OCon(s,ts,t)
    | S.TyProduct(ts) -> Orn.Soup.OProduct(ts)
    | S.TyId(s, l) -> Orn.Soup.OId(s, l)

  type scheme =
      tyvar list * ty
end



module Solver = SolverHi.Make(Id.TeVar)(S)(O)

open Solver

open Ast
module X = Ast.Explicit

let list_diff l1 l2 = List.filter (fun x -> not (List.mem x l2)) l1


let conj_list l =
  List.fold_right (fun x y -> x ^& y >>= fun (x,y) -> x::y) l (pure [])


(* Create a mapping from type variables to bla variables *)
let freshen vars k =
  let rec remap acc k = function
    | [] -> k (List.rev acc) >>= fun x -> ([],x)
    | i::is -> exist (fun v -> remap ((i,v)::acc) k is) >>= fun (x,(xs,y)) -> (x::xs,y)
  in
  remap [] k vars

module E = Eml.ExplicitGml


(* returns a unit constraint *)
let rec id_encode mapping tau : ty -> unit co = function
  | TyVar v ->
    tau --
      (try List.assoc v mapping
       with Not_found -> failwith "unbound variable (id_encode)")
  | TyArrow(t1,t2) ->
    exist_ (fun v1 -> exist_ (fun v2 ->
        id_encode mapping v1 t1 ^^
          id_encode mapping v2 t2 ^^
          tau --- S.TyArrow(v1,v2)
    ))
  | TyProduct(l) ->
    let rec check acc = function
      | [] -> tau --- S.TyProduct (List.rev acc)
      | t::ts -> exist_ (fun v -> id_encode mapping v t ^^ check (v::acc) ts)
                 >>= fun _ -> ()
    in
    check [] l
  | TyCon(c,l) ->
    let rec check acc = function
      | [] -> exist_ (fun v ->
                      tau --- S.TyCon (c, List.rev acc,v) ^^
                        v --- S.TyId (c, List.rev acc) )
      | t::ts -> exist_ (fun v -> id_encode mapping v t ^^ check (v::acc) ts)
                 >>= fun _ -> ()
    in
    check [] l

let rec orn_encode mapping tau : orn -> unit co = function
  | OVar v ->
     tau --
       (try List.assoc v mapping
        with Not_found -> Format.eprintf "%a@." Pp.id v; failwith "unbound variable (orn_encode)")
  | OProduct(l) ->
    let rec check acc = function
      | [] -> tau --- S.TyProduct (List.rev acc)
      | t::ts -> exist_ (fun v -> orn_encode mapping v t ^^ check (v::acc) ts)
                 >>= fun _ -> ()
    in
    check [] l
  | OIdOrn(c, l) ->
     let rec check acc = function
      | [] -> exist_ (fun v ->
                      tau --- S.TyCon (c, List.rev acc,v) ^^
                        v --- S.TyId (c, List.rev acc) )
      | t::ts -> exist_ (fun v -> orn_encode mapping v t ^^ check (v::acc) ts)
                 >>= fun _ -> ()
    in
    check [] l
  | OBaseOrn _ -> failwith "ornament type not generic"
  | OArrow(t1, t2) ->
    exist_ (fun v1 -> exist_ (fun v2 ->
        orn_encode mapping v1 t1 ^^
          orn_encode mapping v2 t2 ^^
          tau --- S.TyArrow(v1,v2)
    ))

let encode_constraint mapping c ws v =
  let rec check acc = function
      | [] -> exist_ (fun phi ->
                        v --- S.TyCon (c, List.rev acc, phi) )
      | t::ts -> exist_ (fun v -> orn_encode mapping v t ^^ check (v::acc) ts)
                 >>= fun _ -> ()
    in
    check [] ws

let encode_scheme osc v =
  (* Introduce a mapping of the variables *)
  let vars = osc.ornscheme_paravars @ List.map (fun (v,_,_) -> v) osc.ornscheme_ornvars in
  let rec build acc = function
    | [] ->
       conj_list (List.map (fun (v,c,ws) -> encode_constraint acc c ws (List.assoc v acc)) osc.ornscheme_ornvars)
       ^^ orn_encode acc v osc.ornscheme_ornty
    | x::xs -> exist_ (fun v -> build ((x,v)::acc) xs)
  in
  build [] vars

let decode_orn t = unsoup t
let gen_of_orn t =
  let rec dec = function
    | OIdOrn(c, ws) -> E.TCon(c, List.map dec ws)
    | OVar(i) -> E.TVar(i)
    | OArrow(w1, w2) -> E.TArr(dec w1, dec w2)
    | _ -> failwith "Not supported in inference"
  in
  dec t
let decode_gen t = gen_of_orn (decode_orn t)


let decode_skel genv = function
  | Soup.OCon(c,ws,inst) | Soup.OCycle(_,Soup.OCon(c,ws,inst))
                           -> E.TCon(List.assoc c genv.skel_map, List.map decode_gen ws)
  | _ -> failwith "not a constructor"


(* Goal of the inference: return \Gamma, \omega, a+ .
   Be careful to: generate good types in a+
*)
let rec infer ?(extra_var = []) (genv : generic_env) (t : X.expr) =
  let instance_requests = ref [] in
  let add_instance_request x = instance_requests := x :: !instance_requests in
  let ident_id = ref 0 in
  let patch_id = ident_id in
  let patch_instances = IdHashtbl.create 43 in
  let introduce_instance loc x paravars ornvars =
    let n = fresh_id x in
    incr ident_id;
    add_instance_request (InstantiateEnv(loc,!ident_id,n,x,List.map decode_orn paravars, List.map decode_orn ornvars));
    n
  in
  let introduce_patch scope loc conid env ret =
    try IdHashtbl.find patch_instances conid
    with Not_found ->
      incr patch_id;
      let patch = id "patch" in
      IdHashtbl.add patch_instances conid (E.Var(patch), ret, dummy_loc);
      add_instance_request (InstantiatePatch(scope,loc,!patch_id,[patch],ret));
      E.Var(patch), ret, dummy_loc
  in
  let introduce_ornvar loc t_rep =
    let intro = function
      | Soup.OCycle(x,Soup.OCon(c,ws, Soup.OVar(v))) ->
         (* That's not subtle at all *)
         let ext, con, proj = id "ext", id "con", id "proj" in
         incr ident_id;
         let ws = List.map (soup_subst_var x v) ws in
         add_instance_request (InstantiateOrnVar(loc, !ident_id, v, c, List.map decode_orn ws, ext, con, proj));
         (ext,con,proj)
      | Soup.OCon(c,ws, Soup.OVar(v)) ->
         (* That's not subtle at all *)
         let ext, con, proj = id "ext", id "con", id "proj" in
         incr ident_id;
         add_instance_request (InstantiateOrnVar(loc, !ident_id, v, c, List.map decode_orn ws, ext, con, proj));
         (ext,con,proj)
      (* Some ornament *)
      | Soup.OCon(c,ws, Soup.OId(c', ws'))
      | Soup.OCycle(_,Soup.OCon(c,ws, Soup.OId(c', ws'))) (* A cycle here is a bit strange... *) ->
         let ext, con, proj = id "ext", id "con", id "proj" in
         add_instance_request (IdentityOrnament(c, List.map decode_orn ws, ext, con, proj));
         (ext, con, proj)
      (* Identity ornament ! *)
      | Soup.OCon(_,_,_) -> failwith "I somehow found another OCon??"
      | _ -> failwith "Not a base ornament"
    in
    let (ext,con,proj) = intro t_rep in
    let phi = decode_gen t_rep in
    let t_skel = decode_gen t_rep in
    let x = id "x" in
    (E.TVar(ext),
     (E.Var(con), E.TGArr(x,t_skel, E.TGArr(id "_", E.TApp( E.TVar(ext), (E.Var x, t_skel, dummy_loc)) , phi)) , dummy_loc),
     (E.Var(proj), E.TGArr(id "_", phi, t_skel) , dummy_loc))
  in

  let instantiate_scheme osc v =
    (* Introduce a mapping of the variables *)
    let rec build acc = function
      | [], [] -> conj_list (List.map (fun (v,c,ws) -> encode_constraint acc c ws (List.assoc v acc)) osc.ornscheme_ornvars)
                  ^^ orn_encode acc v osc.ornscheme_ornty ^^ pure ([], [])
      | x::xs, l -> exist (fun v -> build ((x,v)::acc) (xs,l)) >>= fun (v,(paraty,ornty)) -> (v::paraty, ornty)
      | [], (x,c,ws)::xs ->
         exist (fun v -> build ((x,v)::acc) ([],xs)) >>= fun (v,(paraty,ornty)) -> ignore (introduce_ornvar dummy_loc v); (paraty, v::ornty)
    in
    build [] (osc.ornscheme_paravars, osc.ornscheme_ornvars)
  in

  let rec hastype scope (t, _, loc) tau =
    exist (fun tau' -> tau -- tau' ^^ hastype' scope loc t tau) >>=
      fun (tau_rep, e) -> (e, decode_gen tau_rep, loc)
  and hastype' scope loc (t : X.expr') (tau : variable) : E.expr' co =
    match t with
    | Absurd -> pure E.Absurd
    | X.Var(x, _) when x.id_name = "patch" || x.id_name = "_absurd_" ->
    let patch, _, _ = introduce_patch scope loc x [] (TVar (id "dummy")) in
    pure patch (* XXX find better names *)
    | X.Var(x, _) when not (List.mem x scope) -> Format.eprintf "Scope error on %a@." Pp.id x; failwith "Scope error..."
    | X.Var(x, _) when List.mem_assoc x genv.gen_varenv ->
       let sc = List.assoc x genv.gen_varenv in
       (* That's maybe the only point where we'll use lets... *)
       instantiate_scheme sc tau >>=
         fun (paravars, ornvars) ->
         (* And now we request an instanciation at these specific taus *)
         let x = introduce_instance loc x paravars ornvars in
         E.Var(x)
    | X.Var(x, []) ->
       instance x tau >>=
         fun taus ->
         assert (taus = []);
         E.Var(x)
    | X.Var(x, _) -> pure (E.Var(x)) (*); Format.eprintf "%a@." Pp.id x; failwith "polymorphism unsupported"*)
    | X.App(t1, t2) ->
       exist (fun v ->
        lift (hastype scope) t1 (S.TyArrow(v,tau)) ^&
          hastype scope t2 v
      ) >>= fun (tau,(t1',t2')) ->
      E.App(t1', t2')
    | X.Abs(x,_,u) ->
      let scope = x::scope in
      exist (fun v1 ->
        exist (fun v2 ->
          tau --- (S.TyArrow(v1,v2)) ^&
            def x v1 (hastype scope u v2)
        )
      ) >>= fun (tau1, (tau2, ((), u'))) ->
      E.Abs(x,decode_gen tau1,u')
    (* Non-recursive Let and MultiFix. These let do not generalize. *)
    | X.Let(ismeta, false, [], bindings, e) ->
       let inner_scope = List.map (fun (x,_,_) -> x) bindings @ scope in
       let inner = hastype inner_scope e tau in
       List.fold_right
         (fun (x,_,e') inner_constraint ->
          exist_ (fun vx ->
                  hastype scope e' vx ^&
                    def x vx inner_constraint
                  >>= fun ((_,ty,_) as e', (bs, inner)) -> (x,ty,e')::bs, inner)
         ) bindings (pure [] ^& inner)
       >>= fun (bs, inner) -> E.Let(ismeta, bs, inner)
    | X.Let _ -> failwith "invalid let"
    | X.MultiFix(bindings, ret) ->
       (* Create a variable for each binding, def them,
          typecheck each definition and examine the rest *)
       let fix_scope = List.map (fun (x,_,_) -> x) bindings @ scope in
       let rec do_def = function
         | (x,_,_)::bs -> exist_ (fun v -> def x v (do_def bs))
         | [] ->
            instance ret tau ^^
            conj_list (List.map (fun (x,(y,_),e) ->
                                 pure (x,y) ^& exist (fun v1 -> def y v1
                                  (exist_ (fun v2 ->
                                   exist_ (fun v ->
                                     instance x v ^^
                                     v --- S.TyArrow(v1, v2) ^^
                                     hastype (y::fix_scope) e v2
                                     ))))) bindings)
             >>=
              fun bindings ->
              E.MultiFix(List.map (fun ((x,y),(y_ty, e)) -> (x, (y, decode_gen y_ty), e)) bindings,
                         ret)
       in
       do_def bindings
    (* ShallowCon, ShallowMatch: where we do ornamentation *)
    | X.ShallowCon(c, _, xs,conid) ->
       (* We want to generate only one patch by ornamented conid *)
       exist (fun tau' -> tau -- tau' ^^ con_skel c xs tau) >>=
         fun (tau_rep, (c_skel, xts)) ->
         let x = id "skel" in
         let y = id "patchvar" in
         let phi = decode_gen tau_rep in
         let tau_skel = decode_skel genv tau_rep in
         let E.TCon(_, skel_args) = tau_skel in
         let (phi_ext, phi_con, phi_proj) = introduce_ornvar loc tau_rep in
         (* Well, normally I should also enumerate the scope there... *)
         let patch = introduce_patch scope loc (id "dummy") (*conid*) [] (E.TApp(phi_ext, (E.Var x, tau_skel, loc))) in
         let expr_inj =
           (E.GApp(phi_con,
                   (E.Var x, tau_skel, loc)),
            E.TGArr(id "_",E.TApp(phi_ext, (E.Var x, tau_skel, loc)),phi) , dummy_loc) in
         let xs = List.map (fun (x,ty) -> (E.Var(x), decode_gen ty, dummy_loc)) xts in
         E.Let(IsMeta, [x, tau_skel, (E.Con(c_skel,skel_args,xs), tau_skel, loc)],
               (E.Let(IsMeta, [y, E.TApp(phi_ext, (E.Var x, tau_skel, loc)),
                       patch (* Normally, we should *call* the patch, but... *)
                      ],
                      (E.GApp(expr_inj,
                              (E.Var y, E.TApp(phi_ext, (E.Var x, tau_skel, loc)), dummy_loc))
                      , phi, loc)
                     ), phi, loc))
    | X.ShallowMatch(x, clauses, abs) ->
      (*exist (fun v ->
        hastype scope abs v) >>= fun (tau'', abs) ->*)
       exist (fun tau' ->
       exist (fun tau0 ->
         hastype scope abs tau ^&
               instance x tau0 ^^
                 conj_list (List.map (fun t -> clause_skel scope t tau0 tau) clauses)))
       >>=
         fun ((tau_rep, (tau0_rep, (abs, clauses)))) ->
         let x' = id "skel" in
         (*let y = id "patchvar" in*)
         let phi = decode_gen tau0_rep in
         let tau = decode_gen tau_rep in
         let tau0_skel = decode_skel genv tau0_rep in
         let (phi_ext, phi_con, phi_proj) = introduce_ornvar loc tau0_rep in
         let patch = introduce_patch scope loc (id "dummy") [] (E.TApp(phi_ext, (E.Var x, tau0_skel, loc))) in
         let expr_match = (E.Match( (E.Var(x'), tau0_skel, dummy_loc) , clauses, abs), tau, dummy_loc) in
         let expr_proj = (E.GApp(phi_proj, (E.Var(x), phi, dummy_loc)), E.TGArr(id "_", tau0_skel, phi), dummy_loc) in
         (*let ty_y = E.TApp(phi_ext, (E.Var x, tau0_skel, loc)) in*)
         (*let expr_proj_patched = E.Let(IsMeta, [y, ty_y, patch], (E.GApp(expr_proj, (E.Var y, ty_y, dummy_loc)), tau0_skel, dummy_loc)), tau0_skel, loc in*)
         let expr_proj_patched = (E.GApp(expr_proj, patch), tau0_skel, dummy_loc) in
         E.Let(IsMeta, [x', tau0_skel, expr_proj_patched],
               expr_match)
    | X.Mark(NiceName s, e) -> hastype scope e tau >>= fun e -> E.Label(s,e)
    | X.Mark(m,e) -> hastype scope e tau >>= fun e -> E.Mark(m, e)
    | X.Con _ -> failwith "Unsupported: Con"
    | X.Match _ -> failwith "Unsupported: Match"
    | X.Patch _ ->
      let patch, _, _ = introduce_patch scope loc (id "patch") [] (TVar (id "dummy")) in
      pure patch
  and clause_skel scope (c,xs,e) tau_in tau_out =
    (* Introduce the xs, match tau_in ~ con, elaborate the constructor into a skel-constructor,
       elaborate the action *)
    let rec intro = function
      | [] -> con_skel c xs tau_in ^& hastype (xs@scope) e tau_out
      | x::xs -> exist_ (fun v -> def x v (intro xs))
    in intro xs
       >>=
         fun ((c', map), e) ->
         (c', List.map fst map, e)
  and con_skel c xs tau =
    (* we have a construction c(...).
       we have the ornament type tau = (w_i) t <| phi
       we type c^(...) at (w_i) t^ (by looking at each variable in ...)
     ---
      we return a list (x,ty)s for the xs we were given,
      and we constrain tau' (the type of the skeleton) and phi
     *)
    let c' = try List.assoc c genv.skel_map with | Not_found -> c in
    let tname' = List.assoc c' genv.skel_conenv in
    let tname'_desc = List.assoc tname' genv.skel_dataenv in
    let c'_args = List.assoc c' tname'_desc.skel_conses in
    let tname = List.assoc tname' genv.skel_unmap in
    freshen tname'_desc.skel_args (fun mapping ->
      exist_ (fun phi ->
        tau --- S.TyCon(tname, List.map snd mapping, phi) ^^
          let rec quant args xs = match args, xs with
            | [], [] -> pure []
            | a::args, x::xs -> exist (fun v -> id_encode mapping v a ^^ instance x v ^^ quant args xs) >>= fun (t,bs) -> ((x,t)::bs)
            | _ -> failwith "arg length mismatch"
          in
          quant c'_args xs
            )
    ) >>= fun (_,map) -> (c', map)
  in
  let make_scheme free_vars instreqs oty =
    let ornvars = List.concat (List.map (function
      | InstantiateOrnVar(_,n,v,c,ws,_,_,_) ->
         debug "We have a constraint #%d: %a <| %a (%a)@." n Pp.id v Pp.id c (Pp.seq "," Pp.oty) ws;
         [v, (c,ws)]
      | _ -> []
    ) instreqs)
    in
    (*    let ornvars = List.concat (List.map (fun v -> try let (c,ws) = List.assoc v ornvars in [(v,c,ws)] with Not_found -> []) free_vars) in *)
    let free_vars =
        IdSet.elements (IdSet.of_list (free_vars @ List.concat (List.map (fun (_,(_,ws)) -> List.concat (List.map (fun ot -> IdSet.elements (oty_free_vars ot)) ws)) ornvars))) in
    let paravars = List.filter (fun x -> not (List.mem_assoc x ornvars)) free_vars in
    let ornvars = List.sort_uniq compare (List.map (fun (v,(c,ws)) -> (v,c,ws)) ornvars) in
    {
      ornscheme_paravars = paravars;
      ornscheme_ornvars = ornvars;
      ornscheme_ornty = oty
    }
  in
  try
    let initial_scope = List.map fst genv.gen_varenv @ extra_var in
    solve true ((let0
                    (let1 (id "_toplevel")
                          (fun v -> hastype initial_scope t v)
                          (pure ())
                    )
                 )
                 >>=
                   (* Here we do the work to recover the constraints.
                      Good luck ! *)
                   (fun (_,(free_vars, t_gen, (scheme_vars, oty), _)) ->
                   let instreqs = !instance_requests in
                   make_scheme scheme_vars instreqs (decode_orn oty), (free_vars, instreqs,t_gen))
                )
  with
    | Unify(t1, t2) -> failwith "should never happen"

(* Orn inference:
  * should take the environment into account !
  * should be able to skeletize datatypes
*)

(* We elaborate into a definition of the skeleton, and fill the generic_env *)
let id_skeleton {id_name} = good_id (id_name^"_skel")

(* The definitions here are potentially mutually recursive, except that doesn't do anything for us, so we can simply ignore the problem *)
let elaborate_tydef env genv (tyname,args, conses) =
  let tyname_s = id_skeleton tyname in
  let map_s, consinfo =
    List.split
      (List.map (fun (cname, cargs) ->
                 let cname_s = id_skeleton cname in
                 let cargs_s = List.mapi (fun idx t -> (good_id ("arg_"^cname.id_name^string_of_int idx), t)) cargs in
                 (cname, cname_s),
                 ((cname_s, List.map (fun (x,_) -> TyVar x) cargs_s),
                 cargs_s)
                ) conses)
  in
  let conses_s, mappings_s = List.split consinfo in
  let mapping_s = List.concat mappings_s in
  let args_s = List.map fst mapping_s in
  let map_s = (tyname, tyname_s) :: map_s in
  let unmap_s = List.map (fun (x,y) -> (y,x)) map_s in
  let skel_entry = {
    skel_args = args_s;
    skel_id_args = args;
    skel_id_inst = List.map snd mapping_s;
    skel_conses = conses_s
  } in
  let dataenv_entry = (tyname_s, (args_s, conses_s)) in
  let conenv_entries = List.map (fun (c_s,cargs) -> (c_s, (tyname_s, args_s, cargs))) conses_s in
  {env with
    conenv =  conenv_entries @ env.conenv;
    dataenv = dataenv_entry :: env.dataenv
  },
  {genv with
    skel_map = map_s @ genv.skel_map;
    skel_unmap = unmap_s @ genv.skel_unmap;
    skel_dataenv = (tyname_s,skel_entry) :: genv.skel_dataenv;
    skel_conenv = (List.map (fun (c_s,(tyname_s,_,_)) -> (c_s, tyname_s)) conenv_entries) @ genv.skel_conenv
  },
  (tyname_s, args_s, conses_s)

let elaborate_tydefs env genv (X.TyDef(_,tydefs)) =
  let env, genv, tydefs_s =
    List.fold_right (fun tydef (env,genv,tydefs_s) ->
                     let env,genv,tydef_s = elaborate_tydef env genv tydef in
                     (env,genv,tydef_s::tydefs_s)) tydefs (env,genv,[]) in
  env, genv, X.TyDef(false, tydefs_s)

(* In the no-unfolding case, we treat this exactly as a normal let except it's toplevel.
   In particular, lifting must be done as a mutually recursive group of the same type.
   We'll quickly drop this to allow unfolding optionally, but this is really simpler for now.
   ("quickly drop this" -> we will use the unfolded representation instead)
 *)
let elaborate_valdef genv ?(extra_var=[]) (X.TopLet(isrec, bs)) =
  (* Hum... ok so the no-unfolding is confusing but I see how it could work and I really
     should not split this from the unfolding case. unfolding-or-not will be handled
     at the lifting site, it's much simpler *)
  (* So, what we are doing here is that we are typechecking each binding independently,
     with the self- and mutual-reference as abstract external function.
     That will be considered a _parametric_ constraint I guess. Hope this does not
     mess things up down the road *)
  (* We are ignoring the "isrec" annotation. This has already been handled upstream *)
  let rn_rec = List.map (fun (x,_,_) -> (x, fresh_id x)) bs in
  let varenv_rec = List.map (fun (x,_,_) -> let a = id "a" in (List.assoc x rn_rec, {ornscheme_paravars = [a]; ornscheme_ornvars = []; ornscheme_ornty = OVar a})) bs in
  let derec_renamer, _, _ = SubstRename.(make_subst_full (fun x vs -> try X.Var(List.assoc x rn_rec, [TyVar (id "dummy")]) with Not_found -> X.Var(x,vs)) id_type) in
  let bs_rec = List.map (fun (x,sc,e) -> (x,sc,derec_renamer e)) bs in
  let genv_def = if isrec then {genv with gen_varenv = varenv_rec @ genv.gen_varenv} else genv in
  (* Let's do it and we'll solve the mess later *)
  let elabs = List.map (fun (x,_,e) ->
    let osc, elab = infer ~extra_var genv_def e in
    (x, osc, elab)
  ) bs_rec in
  {genv
   with gen_varenv = List.map (fun (x,osc,_) -> (x,osc)) elabs @ genv.gen_varenv },
  elabs,
  rn_rec

let scheme_to_oscheme (Scheme(vs,t)) =
  let rec tr = function
    | TyArrow(t1, t2) -> OArrow(tr t1, tr t2)
    | TyProduct(ts) -> OProduct(List.map tr ts)
    | TyVar(v) -> OVar(v)
    | TyCon(c,ts) -> OIdOrn(c,List.map tr ts)
  in
  {
    ornscheme_paravars = vs;
    ornscheme_ornvars = [];
    ornscheme_ornty = tr t;
  }

let elaborate_abstract_def genenv x sc =
  {genenv with gen_varenv = (x, scheme_to_oscheme sc) :: genenv.gen_varenv }
