open Util

module Make(O : sig val oenv : Ast.Explicit.orn_env end) = struct
  module S = struct
    type 'a structure =
      | TyArrow of 'a * 'a
      | TyOrn of Ast.id * 'a list

    let map f t =
      match t with
      | TyArrow (t1, t2) ->
         let t1 = f t1 in
         let t2 = f t2 in
         TyArrow (t1, t2)
      | TyOrn(s, l) -> TyOrn(s, List.map f l)


    let fold f t accu =
      match t with
      | TyArrow (t1, t2) ->
         let accu = f t1 accu in
         let accu = f t2 accu in
         accu
      | TyOrn(_, l) -> List.fold_left (fun x y -> f y x) accu l

    let iter f t =
      let _ = map f t in
      ()

    exception Iter2

    let iter2 f t u =
      match t, u with
      | TyArrow (t1, t2), TyArrow (u1, u2) ->
         f t1 u1;
        f t2 u2
      | TyOrn(s1, l1), TyOrn(s2,l2) when s1 = s2 ->
         (try List.iter2 f l1 l2
          with _ when List.length l1 <> List.length l2 -> raise Iter2)
      | _, _ ->
         raise Iter2

    let show sub ppf = function
      | TyArrow(t1,t2) -> Format.fprintf ppf "%a -> %a" sub t1 sub t2
      | TyOrn(s, l) -> Format.fprintf ppf "(%a(%a))" Pp.id s (Pp.seq ", " sub) l
  end

type 'a structure = 'a S.structure
type 'a skeleton = Skeleton of Id.id * 'a list

(* -------------------------------------------------------------------------- *)

(* The data structure maintained by the unifier is as follows. *)

(* A unifier variable is a point of the union-find algorithm. *)

type variable =
    descriptor TUnionFind.point

(* Every equivalence class carries a descriptor which contains the following
   information. *)

(* Some of the fields below are mutable, because our client sometimes needs to
   update them. However, this is never done by the unifier itself, hence never
   done during unification. The unification algorithm is transactional: it
   writes only [TRef]s, so that all changes can be rolled back if unification
   fails. *)

and descriptor = {

  (* Every equivalence class carries a globally unique identifier. When
     a new equivalence class is created, a fresh identifier is chosen,
     and when two classes are merged, one of the two identifiers is kept.
     This identifier can be used as a key in a hash table. One should be
     aware, though, that identifiers are stable only as long as no unions
     are performed. *)

  id : Id.id;

  (* Every equivalence class carries a structure, which is either [None],
     which means that the variable is just that, a variable; or [Some t],
     which means that the variable represents (has been equated with) the
     type [t]. *)

  structure : variable structure option;

  (* This is specific to ornaments: a variable can have a constrained
     skeleton but no structure. As this is directly computed from the
     structure, there is no need to unify it if we unify two variables
     that already have structure. *)

  skeleton : variable skeleton option;
}

(* -------------------------------------------------------------------------- *)

(* Accessors. *)

let id v =
  (TUnionFind.find v).id

let structure v =
  (TUnionFind.find v).structure

(* -------------------------------------------------------------------------- *)

(* [r++]. *)

let postincrement r =
  let v = !r in
  r := v + 1;
  v

(* -------------------------------------------------------------------------- *)

(* [fresh] creates a fresh variable with specified structure and rank. *)

let fresh =
  fun structure ->
    TUnionFind.fresh {
      id = Id.id "a";
      structure = structure;
      skeleton = None;
    }

let fresh_skel =
  fun skel ->
    TUnionFind.fresh {
      id = Id.id "o";
      structure = None;
      skeleton = Some skel
    }
(* -------------------------------------------------------------------------- *)

exception Unify of variable * variable
exception NotABaseOrnament of variable

(* Important invariant: no variable here is created without structure *)
(* This is quite dependent on the Ast... *)
let rec encode_ty map = function
  | Ast.TyVar x -> Id.IdMap.find x map
  | Ast.TyArrow(t1, t2) -> fresh (Some(S.TyArrow(encode_ty map t1, encode_ty map t2)))
  | Ast.TyCon(i,ts) -> fresh (Some(S.TyOrn(i, List.map (encode_ty map) ts)))
  | Ast.TyProduct _ -> Error.failwith "not applicable"

let compute_skeleton = function
  | S.TyArrow _ -> None
  | S.TyOrn(o, ts) ->
     let open Ast in
     let module X = Ast.Explicit in
     let o_desc, _, _, _ = List.assoc o O.oenv.X.env_ornaments in
     match o_desc.X.orn_from_skel with
     | Scheme(free_skel_variables, Ast.TyCon(i_skel, skel_args)) ->
       let ts = List.map (fun _ -> fresh None) free_skel_variables  @ ts in
       let varmap = List.fold_left2 (fun m x v -> IdMap.add x v m) IdMap.empty
                      (X.get_all_orn_args o_desc) ts in
       Some(Skeleton(i_skel, List.map (encode_ty varmap) (skel_args)))
     | _ -> assert false

(* The internal function [unify t v1 v2] equates the variables [v1] and [v2]
   and propagates the consequences of this equation until an inconsistency is
   found or a solved form is reached. In the former case, [S.Iter2] is
   raised. The parameter [t] is a transaction. *)

let rec unify (t : _ TRef.transaction) (v1 : variable) (v2 : variable) : unit =

  (* If the two variables already belong to the same equivalence class, there
     is nothing to do, and [unify_descriptors] is not invoked. Furthermore, if
     there is something to do, then [unify_descriptors] is invoked only after
     the two equivalence classes have been merged. This is not just an
     optimization; it is essential in guaranteeing termination, since we are
     dealing with potentially cyclic structures. *)

  TUnionFind.union_special t preunify_descriptors (unify_descriptors t v1 v2) v1 v2

(* -------------------------------------------------------------------------- *)

(* [unify_descriptors desc1 desc2] combines the descriptors [desc1] and
   [desc2], producing a descriptor for the merged equivalence class. *)
(* _When_ visiting something, it is important to mark it as structured, so we don't
   generate the skeleton for every structured thing we unify it with.
   Unification of the structures already implies that the skeleton are the same,
   because the relation is functional *)
and preunify_descriptors desc1 desc2 =
  { id = desc1.id;
    structure = Option.multiply (fun d1 _ -> d1) desc1.structure desc2.structure;
    skeleton = None; (* Playing with fire *)
  }
and unify_descriptors t v1 v2 desc1 desc2 =
  {
  (* An arbitrary choice of identifier is ok. *)
  id        = desc1.id;
  structure = unify_structures t desc1.structure desc2.structure;
  skeleton = match desc1.structure, desc2.structure with
  | None, None ->
     (* We do not trigger any unfolding *)
     Option.multiply (fun d1 d2 -> unify_skeletons t d1 d2; d1) desc1.skeleton desc2.skeleton
  | Some u, None ->
     begin match compute_skeleton u, desc2.skeleton with
     | None, Some _ -> raise (NotABaseOrnament v1)
     | _, None -> ()
     | Some s1, Some s2 -> ignore (unify_skeletons t s1 s2)
     end;
    None
  | None, Some u ->
     begin match compute_skeleton u, desc1.skeleton with
     | None, Some _ -> raise (NotABaseOrnament v1)
     | _, None -> ()
     | Some s1, Some s2 -> ignore (unify_skeletons t s1 s2)
     end;
    None
  | Some _, Some _ -> None
}
  
(* -------------------------------------------------------------------------- *)

(* [unify_structures structure1 structure2] combines two structures. If one of
   them is undefined, we just keep the other. If both are defined, we unify
   them recursively. *)

and unify_structures t structure1 structure2 =
  Option.multiply (fun t1 t2 ->
    S.iter2 (unify t) t1 t2;
    t2 (* arbitrary; [t1] and [t2] are now equal anyway *)    
  ) structure1 structure2

and unify_skeletons t (Skeleton(i1,vs1)) (Skeleton(i2,vs2)) =
  if i1 <> i2 || List.length vs1 <> List.length vs2 then raise S.Iter2
  else List.iter2 (unify t) vs1 vs2
(* -------------------------------------------------------------------------- *)
(* We expose [unify] as [unify_transactionally] because it is quite useful. *)
let unify_transactionally = unify

(* The public version of [unify] wraps the unification process in a
   transaction, so that a failed unification attempt does not alter the state
   of the unifier. *)

let unify v1 v2 =
  try
    TRef.tentatively (fun t ->
      unify t v1 v2
    )
  with S.Iter2 ->
    raise (Unify (v1, v2))

(* -------------------------------------------------------------------------- *)

module VarMap =
  Hashtbl.Make(struct
    type t = variable
    let equal = TUnionFind.equivalent
    let hash v = Id.IdHash.hash (id v)
  end)

(* -------------------------------------------------------------------------- *)

let equivalent =
  TUnionFind.equivalent

let is_representative =
  TUnionFind.is_representative

(* -------------------------------------------------------------------------- *)

(* The occurs check, which detects cycles in the graph, cannot be defined as
   an instance of the cyclic decoder, for several reasons. For one thing, the
   cyclic decoder is inefficient, as it does not (cannot) mark the nodes that
   have been visited. Furthermore, the occurs check explores only the young
   generation, whereas the decoders explore everything. *)

exception Cycle of variable

let new_occurs_check (is_young : variable -> bool) =

  (* This hash table records the variables that are being visited (they are
     mapped to [false]) or have been visited (they are mapped to [true]). *)

  let table : bool VarMap.t = VarMap.create 128 in

  let rec traverse v =
    if is_young v then
      try
        let visited = VarMap.find table v in
        if not visited then
          (* This node is in the table, but has not been fully visited.
             Hence, it is being visited. A cycle has been found. *)
          raise (Cycle v)
      with Not_found ->
        (* Mark this variable as being visited. *)
        VarMap.add table v false;
        (* Visit its successors. *)
        Option.iter (S.iter traverse) (structure v);
        (* Mark this variable as fully visited. *)
        VarMap.replace table v true
  in

  traverse

(* -------------------------------------------------------------------------- *)

(* Bottom-up computation over an acyclic graph. *)

let new_acyclic_decoder
  (leaf :     variable -> 'a)
  (fold : 'a structure -> 'a)
        :     variable -> 'a =

  (* This hash table records the variables that have been visited and the
     value that has been computed for them. *)

  let visited : 'a VarMap.t = VarMap.create 128 in

  let rec decode v =
    try
      VarMap.find visited v
    with Not_found ->
      (* Because the graph is assumed to be acyclic, it is ok to update
         the hash table only after the recursive call. *)
      let a = 
        match structure v with
        | None ->
            leaf v
        | Some t ->
            fold (S.map decode t)
      in
      VarMap.add visited v a;
      a

  in
  decode

(* -------------------------------------------------------------------------- *)

(* The cyclic decoder is designed mainly with the goal of constructing recursive
   types using [\mu] syntax. We must ensure that every use of a [\mu]-bound is
   dominated by its binder. This makes it impossible to use a table of [visited]
   nodes and share their value; we would risk entering an already-visited cycle
   via a different path. In order to avoid this problem, we define a decoder that
   uses a [visiting] table, but no [visited] table. This makes it correct, but
   potentially exponentially inefficient. Use with caution! *)

let new_cyclic_decoder
  (leaf      :     variable -> 'a)
  (fold      : 'a structure -> 'a)
  (temporary :     variable -> 'a)
  (mu        :     'a -> 'a -> 'a)
             :     variable -> 'a =

  (* This hash table records the variables that are being visited and (if
     these variables have been recursively re-discovered) the temporary value
     that has been assigned to them. *)

  let visiting : 'a option VarMap.t = VarMap.create 128 in

  let rec decode v =
    try
      begin match VarMap.find visiting v with
      | None ->
          (* This variable is being recursively re-discovered for the
             first time. Compute a temporary value for it, and update
             the table. *)
          let a = temporary v in
          VarMap.replace visiting v (Some a);
          a
      | Some a ->
          (* This variable is being recursively re-discovered again.
             We already have a temporary value for it in the table. *)
          a
      end
    with Not_found ->
      (* This variable is not being visited. It has never been found
         before. Before the recursive call, mark it as being visited. *)
      VarMap.add visiting v None;
      (* Perform the recursive traversal. *)
      let a = 
        match structure v with
        | None ->
            leaf v
        | Some t ->
            fold (S.map decode t)
      in
      (* Mark this variable as no longer being visited. Mark it as visited
         instead. If it was recursively re-discovered during the recursive
         call, then compute the final value as a combination of the temporary
         value and the computed value. *)
      let ta = VarMap.find visiting v in
      VarMap.remove visiting v;
      match ta with
      | None ->
          (* This variable was not recursively hit. *)
          a
      | Some ta ->
          (* This variable was recursively hit. Create a [mu] binder. *)
          mu ta a

  in
  decode

end
