open Ast
open Ast.Explicit

exception Not_unifiable of expr * expr
exception Not_unifiable_pat of pat * pat

type zipper = expr -> expr

let to_patch (i, e, e') : patch =
  let prepare_expr e' e =
    PrettyNames.run_renamer (fun s ->
        ignore PrettyNames.(intro s NsCons sum_l);
        ignore PrettyNames.(intro s NsCons sum_r);
        let e = Pretty.pretty_term ~new_scope:{scope = fun s f -> f s} s e in
        let e' = Pretty.pretty_term ~new_scope:{scope = fun s f -> f s} s e' in
        e', e
      )
  in
  let patch_addr = (PatchFor [(Some (UserId "a"), None), None]) in
  let e'_, e_ = prepare_expr e' e in
  let e'_ = Patch.insert_patch e'_ e_ in
  PatchHole(e'_, patch_addr, e_)

let rec find_patch (wanted_vars: IdSet.t) (f: zipper) (subst : (id * id) list) (e : expr) (e' : expr) : (id * expr * expr) list =
  let e_, ty, loc = e in
  let abs = Var (id "__", []), ty, loc in
  let e'_, _, _ = e' in
  let handle_match e e' cls cls' =
    (* patterns in cls are necessarily more complete than patterns in cls'.
       Therefore, the plan is to iterate on cls', find every pattern compatible
       in cls, and then to combine the results *)
    find_patch wanted_vars (fun x ->
        let _, ty_, loc_ = e in
        f (Match (x, [(PWild, ty_, loc_), abs]), ty, loc)) subst e e' @
    (List.map (fun (p', e') ->
         List.map (fun (p, e) ->
             try
               let subst = subst @ unify_patch p p' in
               find_patch wanted_vars (fun x ->
                   f (Match (abs, [p, x]), ty, loc)) subst e e'
             with Not_unifiable_pat (p, p') -> []
           ) cls
         |> List.concat) cls'
     |> List.concat)
  in
  let is_patch i = i.id_name = "patch" || IdSet.mem i wanted_vars in
  match e_, e'_ with
  | _, Mark (_, e') -> find_patch wanted_vars f subst e e'
  | Mark (_, e), _ -> find_patch wanted_vars f subst e e'
  | _, Match ((Var(i, _), _, _) as e'', p_cls) when is_patch i ->
    begin try
        search p_cls (fun (p, cls) ->
            try
              let set_patvars = (IdSet.of_list (patvars p)) in
              let patches = find_patch (IdSet.union set_patvars wanted_vars) f subst e cls in
              let patched_variables, patches = List.partition (fun (v, _, _) -> IdSet.mem v set_patvars) patches in
              let e = List.fold_left (fun e (v, e', _) ->
                  Let(Ast.IsUser, false, [], [v, Scheme([], ty), e'], e), ty, dummy_loc)
                  (Orn_codegen.pat_to_expr ~wild_to_unit:true p)
                  patched_variables in
              let patch_addr = (PatchFor [(Some (UserId "a"), None), None]) in
              let e' =  f (Match ((Patch patch_addr, ty, loc), [(PWild, ty, loc), abs]), ty, loc) in
              Some ([i, e, e'] @ patches)
            with Not_unifiable _ -> None)
      with
      | Not_found -> match e_ with | Match(e, cls) -> handle_match e e'' cls p_cls
                                   | _ -> raise (Not_unifiable (e, e'))
    end
  | _, Var(i', _) when is_patch i' ->
    let patch_addr = (PatchFor [(Some (UserId "a"), None), None]) in
    let e' =  f (Patch patch_addr, ty, loc) in
    [i', e, e']
  | Var (i, _), Var(i', _) ->
    let i =
      try List.assoc i subst
      with
      | Not_found -> i in
    if i <> i' then
      raise (Not_unifiable (e, e'))
    else
      []
  | Let (is_meta, is_rec, _, binds_, e), Let(_, _, _, binds', e') ->
    let binds = List.combine binds_ binds' in
    let subst = subst @ (List.map (fun ((v, _, _), (v', _, _)) -> v, v') binds) in
    (List.map (fun ((v, ty_, e), (_, _, e')) ->
         find_patch wanted_vars (fun x ->
             f (Let(is_meta, is_rec, [], [(v, ty_, x)], abs), ty, loc)
           ) subst e e') binds
     |> List.concat) @
    find_patch wanted_vars (fun x ->
        let binds_ = List.map (fun (v, ty_,_) -> v, ty_, abs) binds_ in
        f (Let(is_meta, is_rec, [], binds_, x), ty, loc)) subst e e'
  | Match (e, cls), Match(e', cls') ->
    handle_match e e' cls cls'
  | Con (con, tys, l), Con(i', _, l') when con = i' ->
    List.combine l l'
    |> List.mapi (fun i (a, b) ->
        find_patch wanted_vars (fun x ->
            f (Con(con, tys, List.mapi (fun j _ -> if i = j then x else abs) l), ty, loc))
          subst a b)
    |> List.concat
  | Abs (v, ty_, e), Abs(v', _, e') ->
    find_patch wanted_vars (fun x -> f (Abs(v, ty_, x), ty, loc))
      ((v, v')::subst) e e'
  | App (a, b), App(a', b') ->
    find_patch wanted_vars (fun x -> f (App(x, abs), ty, loc)) subst a a' @
    find_patch wanted_vars (fun x -> f (App(abs, x), ty, loc)) subst b b'
  | ShallowMatch _, ShallowMatch _ -> assert false
  | ShallowCon _, ShallowCon _ -> assert false
  | Absurd, Absurd  -> assert false
  | Patch _, Patch _ -> assert false
  | MultiFix (l, i), MultiFix (l', i'') ->
    if List.length l <> List.length l' then raise (Not_unifiable (e, e'))
    else
      let subst = subst @ List.map2 (fun (i, _, _) (i', _, _) -> (i, i')) l l' in
      List.map2 (fun (i, (arg, arg_ty), e) (i', (arg', arg_ty'), e') ->
          find_patch wanted_vars (fun x -> f (MultiFix([i', (arg', arg_ty'), x], i''), ty, loc)) ((arg, arg') :: subst) e e') l l'
      |> List.concat
  | _ -> raise (Not_unifiable (e, e'))
and unify_patch (p: pat) (p': pat) =
  let p_, _, _ = p in
  let p'_, _, _ = p' in
  match p_, p'_ with
  | _, PWild -> []
  | PCon(i, _, l), PCon(i', _, l') when i = i' ->
    List.map2 unify_patch l l'
    |> List.concat
  | PVar i, PVar i' -> [i, i']
  | PAlt(a, b), PAlt(a', b') -> unify_patch a a' @ unify_patch b b'
  | PAlt(a, b), _ ->
    begin
      try
        unify_patch a p'
      with Not_unifiable_pat _ -> unify_patch b p'
    end
  | PZero, _ -> assert false
  | _ -> raise (Not_unifiable_pat (p, p'))

let free_vars e: IdSet.t =
  Ast.Explicit.visit_expr (fun () s e ->
      All (), (match e with
          | Var (i, _), _, _ -> IdSet.add i s
          | _ -> s), e) () IdSet.empty e
  |> fst

let defined_vars e: IdSet.t =
  match e with
  | Let(_, _, _, l, _), _, _ -> List.map (fun (v, _, _) -> v) l |> IdSet.of_list
  | Match(_, [p, _]), _, _ ->
    patvars p |> IdSet.of_list
  | _ -> IdSet.empty

let sub_term (i1, patch_e1, e1 : id * expr * expr) (i, patch_e2, e2 : id * expr * expr) =
  let b =
    Ast.Explicit.visit_expr (fun () is_eq e ->
        let l = ref [] in
        let b = Dead_code_elimination.eq_modulo_alpha_aux ~equalize:(fun x -> l := x :: !l) [i1, i] e1 e in
        let b = b && not (Dead_code_elimination.eq_modulo_alpha_aux !l patch_e1 patch_e2) in
        All (), (is_eq || b), e) () false e2
    |> fst in
  b

let minimize (ps : (id * expr * expr) list) : (id * expr * expr) list =
  let ps = List.map (fun (i, e2, e1) ->
      let es = List.filter ( fun (v, _, _) -> v <> i) ps in
      let fv = free_vars e2 in
      i, e2, Ast.Explicit.visit_expr (fun bound s e ->
          if bound then All true, s, e
          else
            match e with
            | Var _, _, _ -> All true, s, e
            | _ ->
              let vd = defined_vars e in
              let is_subterm = List.exists (sub_term (i, e2, e)) es in
              All (not (IdSet.is_empty (IdSet.inter vd fv)) || is_subterm),
              (if is_subterm then s else e),
              e
        ) false e1 e1
             |> fst) ps in
  let ps = List.fold_left (fun ps (i, patch_e, e) ->
      if List.exists (fun (i', _, e') ->
          Dead_code_elimination.eq_modulo_alpha_aux [i, i'] e e') ps then ps
      else (i, patch_e, e) :: ps
    ) [] ps in
  ps

let find_patch (t: toplet) (t': toplet) : patch list =
  try
    let TopLet(_, bindings) = t in
    let TopLet(_, bindings') = t' in
    begin match bindings, bindings' with
    | [i, _, e], [i', _, e'] ->
    find_patch IdSet.empty (fun x -> x) [i, i'] e e'
    |> minimize
    |> List.map to_patch
    |> List.sort_uniq compare
    |> List.filter (function | PatchHole(_, _, (Var (UserId e), _)) when e.[0] = '#' -> false | _ -> true)
    | _ -> Format.eprintf "Patch generation not supported for recursive functions@."; exit (-1)
    end
  with
  | Not_unifiable((_, _, l), _) -> raise (Error.Error(l, "Could not generate patches"))
