open Ast

let fp = Format.fprintf


let rec seq sep sub ppf = function
  | [] -> ()
  | [x] -> fp ppf "%a" sub x
  | x::xs -> fp ppf "%a%s%a" sub x sep (seq sep sub) xs

let id ppf x = fp ppf "%s%s@@%d" x.id_name (if x.id_user then "!" else "") x.id_unique
let tyvar ppf x = fp ppf "'%s@@%d" x.id_name x.id_unique

let user_id ppf = function
  | UserId x -> fp ppf "%s" x
  | FromId x -> id ppf x
let utyvar ppf = function
  | UserId x -> fp ppf "'%s" x
  | FromId x -> tyvar ppf x

let id_nice ppf x = fp ppf "%s" x.id_name
let tyvar_nice ppf x = fp ppf "'%s" x.id_name

let ty_aux id tyvar  =
  let rec ty p ppf t = match t, p with
    | TyArrow(t1,t2), false -> fp ppf "%a -> %a" (ty true) t1 (ty false) t2
    | TyArrow(t1,t2), true ->  fp ppf "(%a -> %a)" (ty true) t1 (ty false) t2
    | TyVar(x),_ -> tyvar ppf x
    | TyProduct [], _ -> fp ppf "unit"
    | TyProduct ts,_ -> fp ppf "(%a)" (seq " * " (ty false)) ts
    | TyCon(c,[]),_ -> fp ppf "%a" id c
    | TyCon(c,[t]), _ -> fp ppf "%a %a" (ty true) t id c
    | TyCon(c,ts),_ -> fp ppf "(%a) %a" (seq ", " (ty false)) ts id c
  in
  ty false

let ty = ty_aux id tyvar
let user_ty : Format.formatter -> user_ty -> unit = ty_aux user_id utyvar
let ty_nice ppf = ty_aux id_nice tyvar_nice ppf
(*
let (iterm,ipat,itop,iprogram) =
  let rec term ppf = function
    | Var(v) -> fp ppf "%a" user_id v
    | Let(isrec,defs,e) ->
      fp ppf "(let %s%a in %a)" (if isrec then "rec " else "")
        (seq " and " letb) defs term e
    | Tuple(es) -> fp ppf "(%a)" (seq ", " term) es
    | App(u,v) -> fp ppf "(%a %a)" term u term v
    | Abs(x,e) -> fp ppf "(fun %a -> %a)" user_id x term e
    | Con(i,[]) -> fp ppf "%a" user_id i
    | Con(i,es) -> fp ppf "(%a(%a))" user_id i (seq ", " term) es
    | Match(e,cs) -> fp ppf "(match %a with %a)" term e (seq " | " clause) cs
  and letb ppf (i,e) = fp ppf "%a = %a" user_id i term e
  and clause ppf (p,e) = fp ppf "%a -> %a" pat p term e
  and pat ppf = function
    | PVar(v) -> fp ppf "%a" user_id v
    | PWild -> fp ppf "_"
    | PTuple(ps) -> fp ppf "(%a)" (seq ", " pat) ps
    | PCon(i,[]) -> fp ppf "%a" user_id i
    | PCon(i,es) -> fp ppf "%a(%a)" user_id i (seq ", " pat) es
  and tydef ppf = function
    | TyDef(tn,targs,tconses) -> fp ppf "%a %a = %a" typle targs user_id tn (seq " | " tycon) tconses
  and typle ppf = function
    | [] -> fp ppf ""
    | [x] -> utyvar ppf x
    | xs -> fp ppf "(%a)" (seq "," utyvar) xs
  and tycon ppf = function
    | (cn,[]) -> fp ppf "%a" user_id cn
    | (cn,xs) -> fp ppf "%a of %a" user_id cn (seq " * " user_ty) xs
  and top ppf = function
    | StVal(x,e) -> fp ppf "let %a = %a" user_id x term e
    | StType(tys) -> fp ppf "type %a" (seq " and " tydef) tys
  and program ppf prog = fp ppf "%a" (seq "\n\n" top) prog
  in
  (term,pat,top,program)
*)
let tyscheme ppf (Scheme(vs,t)) =
  fp ppf "forall %a. %a" (seq ", " tyvar) vs ty t



let patch_addr impl_term (ft: Format.formatter) (p: patch_addr) =
  match p with
  | Ast.DefaultPatch -> Format.fprintf ft "*"
  | Ast.PatchFor l -> Format.fprintf ft "%a" (seq "," (fun ft ((u, i), e) ->
      begin match u with
        | None -> ()
        | Some u -> Format.fprintf ft "#%a" user_id u
      end;
      begin match i with
      | None -> ()
      | Some u -> Format.fprintf ft "#%d" u end;
      begin
        match e with
        | None -> ()
        | Some e -> Format.fprintf ft "[%a]" impl_term e
      end
)) l

let patch_ident (ft: Format.formatter) (i: patch_ident) =
  match i with
  | Ast.PatchIdent u -> Format.fprintf ft "id %a@." user_id u
  | Ast.PatchSpecial u ->  Format.fprintf ft "spec %a@." user_id u



let mk_impl_term_pat id =
  let rec term' ppf t = match t with
    | Absurd -> fp ppf "<absurd>"
    | Patch p -> fp ppf "#[%a]" (patch_addr term) p
    | Var(s) -> fp ppf "%a" id s
    | Let(t, isrec,defs,e) ->
       fp ppf "(let%s %s %a in %a)"
          (if t = IsShare && false then "<share>" else "")
          (if isrec then "rec " else "")
          (seq " and " letb) defs
          term e
    | App(u,v) -> fp ppf "(%a %a)" term u term v
    | Abs(x,e) -> fp ppf "(fun %a -> %a)" id x term e
    | Match(e,cs) -> fp ppf "(match %a with %a)" term e (seq " | " clause) cs
    | Con(i,es) -> fp ppf "%a(%a)" id i (seq ", " term) es
    | Tuple(es) -> fp ppf "(%a)" (seq ", " term) es
    | Label(_,e) -> term ppf e
(*    | Wild -> fp ppf "_" *)
  and letb ppf (i,e) = fp ppf "%a = %a" id i term e
  and clause ppf (p,e) = fp ppf "%a -> %a" pat p term e
  and pat' ppf = function
    | PVar(v) -> fp ppf "%a" id v
    | PWild -> fp ppf "_"
    | PCon(i,ps) -> fp ppf "%a(%a)" id i (seq ", " pat) ps
    | PTuple(ps) -> fp ppf "(%a)" (seq ", " pat) ps
    | PZero -> fp ppf "~"
    | PAlt(p1,p2) -> fp ppf "(%a | %a)" pat p1 pat p2
  and pat ppf (p,_) = pat' ppf p
  and term ppf (e,_) = term' ppf e
  in
  (term,pat)

let impl_term, impl_pat = mk_impl_term_pat user_id

let patch_addr = patch_addr impl_term

let patch (ft: Format.formatter) (p: patch) = match p with
| Ast.PatchHole (a, b, c) -> Format.fprintf ft "patch %a" impl_term a
| Ast.PatchLifting (a, b) -> Format.fprintf ft "lifting %a <- %a" patch_addr a (seq "," patch_ident) b
| Ast.PatchOrnament (a, l) -> Format.fprintf ft "ornament %a <- %a" patch_addr a (seq "," patch_ident) l

let xterm, xpat =
  let open Explicit in
  let rec term' ppf t = match t with
    | Absurd -> fp ppf "<absurd>"
    | Patch p -> fp ppf "%a" patch_addr p
    | Var(v,[]) -> fp ppf "%a" id v
    | Var(v,ts) -> fp ppf "%a@@[%a]" id v (seq ", " ty) ts
    | Let(_,isrec,tyvars,defs,e) ->
      fp ppf "(let %s[%a] %a in %a)" (if isrec then "rec " else "")
        (seq "," tyvar) tyvars
        (seq " and " letb) defs
        term e
    | App(u,v) -> fp ppf "(%a %a)" term u term v
    | Abs(x,t,e) -> fp ppf "(fun (%a : %a) -> %a)" id x ty t term e
    | Match(e,cs) -> fp ppf "(match %a with %a)" term e (seq " | " clause) cs
    | ShallowMatch(x, cs, abs) ->
       fp ppf "(s-match %a with %a | <absurd> -> %a)" id x (seq " | " sclause) cs term abs
    | ShallowCon(i, ts, xs, _conid) ->
       fp ppf "s-%a[%a](%a)" id i (seq "," ty) ts (seq ", " id) xs
    | Mark(m,e) -> fp ppf "%a" term e
    | MultiFix(bindings, sel) ->
       fp ppf "(fix %a in %a)" (seq " and " mfb) bindings id sel
    | Con(i,ts,es) -> fp ppf "%a[%a](%a)" id i (seq "," ty) ts (seq ", " term) es
  and letb ppf (i,Scheme(vs,t),e) = fp ppf "%a : %a = %a" id i ty t term e
  and mfb ppf (x,(y,y_ty),e) = fp ppf "%a (%a : %a) -> %a" id x id y ty y_ty term e
  and clause ppf (p,e) = fp ppf "%a -> %a" pat p term e
  and sclause ppf (c, xs, e) = fp ppf "%a(%a) -> %a" id c (seq "," id) xs term e
  and pat' ppf = function
    | PVar(v) -> fp ppf "%a" id v
    | PWild -> fp ppf "_"
    | PCon(i,ts,ps) -> fp ppf "%a[%a](%a)" id i (seq "," ty) ts (seq ", " pat) ps
    | PZero -> fp ppf "~"
    | PAlt(p1,p2) -> fp ppf "(%a | %a)" pat p1 pat p2
  and pat ppf (p,_,_) = pat' ppf p
  and term ppf (e,_,_) = term' ppf e
  in
  term, pat

let xtop ppf (Explicit.TopLet(isrec, bindings)) =
  let xbind ppf (x,scheme,e) = fp ppf "%a : %a = %a\n" id x tyscheme scheme xterm e in
  fp ppf "let %s%a" (if isrec then "rec " else "") (seq "\nand " xbind) bindings

let mk_term_pat id =
  let open Explicit in
  let rec term' ppf t = match t with
    | Absurd -> fp ppf "<absurd>"
    | Patch _ -> fp ppf "#"
    | Var(s,[]) -> fp ppf "%a" id s
    | Var(s,ts) -> fp ppf "%a" id s
    | Let(t, isrec,tyvars,defs,e) ->
       fp ppf "(let%s %s %a in %a)"
          (if t = IsShare && false then "<share>" else "")
          (if isrec then "rec " else "")
          (seq " and " letb) defs
          term e
    | App(u,v) -> fp ppf "(%a %a)" term u term v
    | Abs(x,t,e) -> fp ppf "(fun %a -> %a)" id x term e
    | Match(e,cs) -> fp ppf "(match %a with %a)" term e (seq " | " clause) cs
    | Con(i,ts,es) -> fp ppf "%a(%a)" id i (seq ", " term) es
    | ShallowMatch(x, cs, abs) ->
       fp ppf "(s-match %a with %a | <absurd> -> %a)" id x (seq " | " sclause) cs term abs
    | ShallowCon(i, ts, xs, _conid) ->
       fp ppf "%a(%a)" id i (seq ", " id) xs
    | Mark(m,e) -> fp ppf "%a" term e
    | MultiFix(bindings, sel) ->
       fp ppf "(fix %a in %a)" (seq " and " mfb) bindings id sel
(*    | Wild -> fp ppf "_" *)
  and letb ppf (i,Scheme(vs,t),e) = fp ppf "%a = %a" id i term e
  and mfb ppf (x,(y,y_ty),e) = fp ppf "%a %a -> %a" id x id y term e
  and clause ppf (p,e) = fp ppf "%a -> %a" pat p term e
  and sclause ppf (c, xs, e) = fp ppf "%a(%a) -> %a" id c (seq "," id) xs term e
  and pat' ppf = function
    | PVar(v) -> fp ppf "%a" id v
    | PWild -> fp ppf "_"
    | PCon(i,ts,ps) -> fp ppf "%a(%a)" id i (seq ", " pat) ps
    | PZero -> fp ppf "~"
    | PAlt(p1,p2) -> fp ppf "(%a | %a)" pat p1 pat p2
  and pat ppf (p,_,_) = pat' ppf p
  and term ppf (e,_,_) = term' ppf e
  in
  (term,pat)

let term, pat = mk_term_pat id
let term_nice, pat_nice = mk_term_pat id_nice

let top ppf (Explicit.TopLet(isrec, bindings)) =
  let bind ppf (x,_,e) = fp ppf "%a = %a\n" id x  xterm e in
  fp ppf "let %s%a" (if isrec then "rec " else "") (seq "\nand " bind) bindings

(*
open Orn

let gty =
  let rec gty p ppf t = match t, p with
    | GTyArrow(t1,t2), false -> fp ppf "%a -> %a" (gty true) t1 (gty false) t2
    | GTyArrow(t1,t2), true ->  fp ppf "(%a -> %a)" (gty true) t1 (gty false) t2
    | GTyVar(x),_ -> tyvar ppf x
    | GTyProduct ts,_ -> fp ppf "(%a)" (seq ", " (gty false)) ts
    | GTyOrn(OrnId s,[]),_ -> fp ppf "=%a" id s
    | GTyOrn(OrnId s,[t]), _ -> fp ppf "%a =%a" (gty true) t id s
    | GTyOrn(OrnId s,ts),_ -> fp ppf "(%a) =%a" (seq ", " (gty false)) ts id s
    | GTyOrn(OrnNamed s,[]),_ -> fp ppf "%a" id s
    | GTyOrn(OrnNamed s,[t]), _ -> fp ppf "%a %a" (gty true) t id s
    | GTyOrn(OrnNamed s,ts),_ -> fp ppf "(%a) %a" (seq ", " (gty false)) ts id s
  in
  gty false

let head ppf = function
  | HeadVar(v) -> fp ppf "'%a" id v
  | HeadOrn(OrnId(i)) -> fp ppf "=%a" id i
  | HeadOrn(OrnNamed(i)) -> id ppf i

let gkind ppf = function
  | KTy -> fp ppf "*"
  | KOrn(a,b,None) -> fp ppf "orn[%a,%a,_]" ty a ty b
  | KOrn(a,b,Some(h)) -> fp ppf "orn[%a,%a,%a]" ty a ty b head h
  | KHead -> fp ppf "#"

let gscheme ppf (GScheme(vars,t)) =
  let quant ppf = fun (v,k) -> fp ppf "%a : %a" id v gkind k in
  fp ppf "forall %a. %a" (seq ", " quant) vars gty t

module X = Orn.Explicit

let rec xopat ppf (e,_,_) =
  match e with
    | X.PCon(_,i,_,ps) -> fp ppf "%a(%a)" id i (seq "," xopat) ps
    | X.PVar(i) -> id ppf i
    | X.PTuple(ps) -> fp ppf "(%a)" (seq "," xopat) ps
    | X.PWild -> fp ppf "_"

let patchty ppf (X.PaTopLet(vars,lty,p)) =
  let pat = xopat in
  let rec pty ppf p = match fst p with
    | X.PaLet(lb,_,vars,binds,p) ->
      fp ppf "%d: let {%a} {%a} in [%a]" lb
        (seq ", " letvar) vars (seq ";\n" letbind) binds (seq ";\n" pty) p
    | X.PaCtx(_,p) -> fp ppf "[%a]" (seq ";\n" pty) p
    | X.PaInst(lb, x, t) -> fp ppf "%d: inst %a @ %a" lb id x gty t
    | X.PaClause(lb,_,u,v) -> fp ppf "CLAUSE %d: [%a] => [%a] END" lb (seq ";\n" papty) u (seq ";\n" pty) v
    | X.PaCon(lb,_,bits,((_,ty,_) as p)) -> fp ppf "%d: %a |- %a : %a" lb (seq ", " idandty) bits  pat p gty ty
  and papty ppf p = match fst p with
    | X.PaPaCon(lb,_,bits,((_,ty,_) as p)) -> fp ppf "%d: %a |- %a : %a" lb (seq ", " idandty) bits pat p gty ty
  and letvar ppf = function
    | (x, lb, k) -> fp ppf "%d: %a : %a" lb tyvar x gkind k
  and letbind ppf = fun (lb,_,name,t,p) ->
    fp ppf "%d: %a : %a => [%a]" lb id name gty t (seq ";\n" pty) p
  and idandty ppf (_,i,t) = fp ppf "%a : %a" id i gty t
  in
  fp ppf "val {%a} : %a => [%a]" (seq ", " letvar) vars gty lty (seq ";\n" pty) p

let nicepatch ppf (name,X.PaTopLet(vars,lty,p)) =
  let rec pat ppf (e,_,_) = match e with
    | X.PCon(_,i,_,ps) -> fp ppf "%a(%a)" id i (seq "," pat) ps
    | X.PVar(i) -> id ppf i
    | X.PTuple(ps) -> fp ppf "(%a)" (seq "," pat) ps
    | X.PWild -> fp ppf "_"
  in
  let rec pty ppf p = match fst p with
    | X.PaLet(lb,_,vars,binds,p) ->
      fp ppf "%d: let [[%a] [%a]] in [%a];\n" lb (seq "" orninst) vars
        (seq ";\n" letbind) binds (seq "" pty) p
    | X.PaCtx(_,p) -> fp ppf "%a" (seq "" pty) p
    | X.PaInst(lb, x, t) -> fp ppf "%d: lifting (* inst of %a *);\n" lb id x
    | X.PaClause(lb,_,u,v) -> fp ppf "%d : match [[%a] -> [%a]];\n"
      lb (seq "; " papty) u (seq "" pty) v
    | X.PaCon(lb,_,_,e) -> fp ppf "%d : ornament (*ornament of %a*);\n"
      lb pat e
  and papty ppf = function
    | X.PaPaCon(lb,_,_,p), _ ->
      fp ppf "%d: (*ornament of %a*)" lb pat p
  and letbind ppf = fun (lb,_,name,t,p) ->
    fp ppf "%d: (*new name for %a*) = [%a]" lb id name (seq "" pty) p
  and orninst ppf = function
    | (_, lb, KOrn(a,_,_))-> fp ppf "%d : (*orn of %a*); " lb ty a
    | _ -> ()
  in
  fp ppf "let (*name*) = nlifting of %a ornament _ where [%a] patch %a\n"
    id name (seq "" orninst) vars (seq "\n" pty) p

    (*
module E = Patch.Explicit
let einst ppf (E.ITop(tyvars, tyinsts, (newname, _, inst))) =
  fp ppf "liftng of _ with\n let %a {%a} [%a] =\n[%a]\n"
    *)
    *)
let loc ppf (path, pos) =
  let open Lexing in
  let bol = pos.lex_start_p.pos_bol in
  let fname = pos.lex_start_p.pos_fname in
  let path = if path = "<dummy>" || fname = "" then path else fname in
  fp ppf "File %S, line %d, characters %d-%d:\n"
    path pos.lex_start_p.pos_lnum
    (pos.lex_start_p.pos_cnum - bol)
    (pos.lex_curr_p.pos_cnum - bol)

type pos = Arg | Expr2 | Branch | Expr
(* Parentheses ? *)
let (piterm,pipat,pitop,piprogram) =
  let open PPrint in
  let user_id = function
    | UserId x -> string x
    | FromId x -> string (x.id_name ^ "@" ^ string_of_int x.id_unique)
  in
  let utyvar v = if v = UserId "_" then user_id v else string "'" ^^user_id v in
  let tuple sub l = group (lparen ^^ (nest 1 (separate_map (string "," ^^ break 1) sub l)) ^^ break 0 ^^ rparen) in
  let con sub subb c l =
    match l with
    | [] -> user_id c
    | [e] -> user_id c ^^ string " " ^^ subb e
    | l -> user_id c ^^ tuple sub l
  in
  let rec term prio t =
    let parif p g =
      group (if p > prio then enclose lparen rparen g else g)
    in
    let beginif p g =
      group (if p > prio then enclose (string "begin ") (hardline ^^ string "end") g else g)
    in
    match fst t with
      | Var(v) -> parif Arg (user_id v)
      | Let(_,isrec,defs,e) ->
        parif (max prio Branch) (group (string (if isrec then "let rec " else "let ") ^^
                               separate_map (hardline ^^ string " and ") letb defs ^/^
                               string "in") ^^ hardline ^^ term prio e)
      | Tuple(es) -> tuple (term Expr) es
      | App(u,v) -> parif Expr2 (term Expr2 u ^/^ (nest 2 (term Arg v)))
      | Abs(x,e) ->
        parif Branch (string "fun "^^user_id x^^string " ->"^//^ term Expr e)
      | Con(i,es) -> parif (if es = [] then Arg else Branch) (con (term Expr) (term Arg) i es)
      | Match(e,cs) ->
        beginif Expr (group (string "match " ^^ term Expr e ^//^ string "with") ^^
                      nest 2 (hardline ^^ separate_map  hardline clause cs))
      | Label(s, e) -> string ("(* "^s^" *) ") ^^ term prio e
      | Patch p -> patch_addr p
      | Absurd -> string "<absurd>"
  and letb (i,e) =
    let rec aux = function
      | Abs(x, (Match( (Var(x'), _) , cs), _)), _
          when x = x' && not (List.exists (free_clause x) cs) && x = UserId "__function__"
        -> string " = function" ^^ (nest 2 (hardline ^^separate_map hardline clause cs))
      | Abs(x,e), _ -> string " " ^^ user_id x ^^ aux e
      | e -> string " = " ^^ (nest 2 (term Expr e))
    in
    group (user_id i ^^ aux e)
  and clause (p,e) =
    group (string "| " ^^pat p ^^ string " ->" ^//^ term Branch e)
  and pat p = match fst p with
    | PVar(v) -> user_id v
    | PWild -> string "_"
    | PTuple(ps) -> tuple pat ps
    | PCon(i,ps) -> con pat pat' i ps
    | PZero -> string "~"
    | PAlt(p1, p2) -> group (enclose lparen rparen (pat p1 ^//^ string "| " ^^ pat p2))
  and pat' p = match fst p with
    | PVar _ | PWild | PTuple _ | PCon(_,[]) | PZero | PAlt _ -> pat p
    | _ -> group (enclose lparen rparen (pat p))
  and tydef = function
    | TyDef(tn,targs,[tcons]) ->
      group (typle utyvar targs ^^ user_id tn ^^ string " =" ^//^
               tycon tcons)
    | TyDef(tn,targs,tconses) ->
      group (typle utyvar targs ^^ user_id tn ^^ string " =" ^//^
               string "| " ^^ separate_map (hardline^^ string "| ") tycon tconses)
  and typle : type a. (a -> _) -> a list -> _ = fun sub -> function
    | [] -> string ""
    | [x] -> sub x ^^ string " "
    | xs -> lparen ^^ separate_map (string ",") sub xs ^^ rparen ^^ string " "
  and tycon = function
    | (cn, []) -> user_id cn
    | (cn,xs) -> group (user_id cn ^^ string " of" ^//^
                          separate_map (string " *" ^^ break 1) (user_ty true) xs)
  and user_ty p =
    let par p' g = if p' < p then lparen ^^ g ^^ rparen else g in
    function
      | TyArrow(t1,t2) -> par false (user_ty true t1 ^^ string " -> " ^^ user_ty false t2)
      | TyVar x -> utyvar x
      | TyProduct [] -> string "unit"
      | TyProduct ts -> lparen ^^ separate_map (string " * ") (user_ty false) ts ^^ rparen
      | TyCon(c,[t]) -> user_ty true t ^^ string " " ^^ user_id c
      | TyCon(c,ts) -> typle (user_ty false) ts ^^ user_id c
  and patch_addr p =
  match p with
  | Ast.DefaultPatch -> string "*"
  | Ast.PatchFor l -> separate_map (string ", ") (function
    | ((Some u, _), None) -> string "#" ^^ user_id u
    | ((Some u, _), Some e) -> string "#" ^^ string "[" ^^ term Expr e ^^ string "]"
    | (_, Some i), None -> string "#" ^^ string (string_of_int i)
      | (_, Some i), Some e -> string "#" ^^ string (string_of_int i) ^^ string "[" ^^ term Expr e ^^ string "]"
    | (None, None), _ -> assert false) l
    and patch_ident p =
    match p with
    | Ast.PatchIdent u -> user_id u
    | Ast.PatchSpecial u -> string "@" ^^ user_id u
  and patch (p, _) = match p with
  | Ast.PatchHole (a, b, c) -> group (string "patch " ^^ term Expr a)
  | Ast.PatchLifting (a, l) -> group (string "lifting " ^^ patch_addr a ^^ separate_map (string ", ") patch_ident l)
  | Ast.PatchOrnament (a, l) ->  group (string "ornament " ^^ patch_addr a ^^ string " <- " ^^ separate_map (string ", ") patch_ident l)
  and orn_clause (a, vars, b) = pat a ^^ string " => " ^^ pat b ^^ match vars with
  | [] -> string ""
  | _ -> string " when " ^^ separate_map (string ",") (fun (a, b) -> user_id a ^^ string ":" ^^ user_ty true b) vars
  and orn_def o =
  group (typle utyvar o.orn_args ^^ user_id o.orn_name ^^ string " : " ^^ user_ty true o.orn_from ^^ string " => " ^^ user_ty true o.orn_to ^^ string " with " ^^ hardline ^^ string " | "
  ^^ separate_map (hardline ^^ string " | ") orn_clause o.orn_clauses)
  (*let _ = match o with | { orn_name; orn_args; orn_from; orn_to; orn_clauses } -> (??) in ??*)
  and top = function
    | StVal(isrec, defs), _ ->
       group (string (if isrec then "let rec " else "let ") ^^
                separate_map (hardline ^^ string " and ") letb defs)
    | StType(tys), _ ->
       group (string "type " ^^ separate_map  (hardline ^^ string "and ") tydef tys)
    | StAbstractVal(x,t), _ ->
       group (string "val " ^^user_id x ^^ string " : " ^//^ user_ty true t)
    | StCommented(isrec, defs), _ ->
      group (string "(*" ^//^ string (if isrec then "let rec " else "let ") ^^
             separate_map (hardline ^^ string " and ") letb defs ^//^ string "*)")
    | StLiftDef l, _ ->
      separate_map hardline (fun l ->
      let lift_at = match l.lift_at with
      | None -> string ""
      | Some s -> string " : " ^^ user_ty true s
      in
      let patches = match l.lift_with with
      | [] -> string ""
      | _ -> string " with" ^^ hardline ^^ string "  "
             ^^ separate_map (hardline^^ string "  ") patch l.lift_with in
      group (string "let " ^^ user_id l.lift_to ^^ string " = lifting " ^^ user_id l.lift_from ^^ lift_at ^^ patches)
        ) l
    | StOrnDef o, _ ->
      group (string "type relation " ^^ separate_map (hardline ^^ string "and") orn_def o)
    | _ -> assert false
  and program prog = separate_map (hardline ^^ hardline) top prog ^^ hardline
  in
  (term,pat,top,program)
(*
let pnicepatch (name,X.PaTopLet(vars,lty,p)) =
  let open PPrint in
  let id (Id(s,_)) = string s in
  let label x = string (string_of_int x) ^^ string ": " in
(*
  let rec unpat = function
    | X.PCon(_,i,_,ps) -> PCon(unid i, List.map unpat ps)
    | X.PVar(i) -> PVar(unpat i)
    | X.PTuple(ps) -> PTuple(List.map unpat ps)
    | X.PWild -> PWild
  in
*)
  let rec pty p = match fst p with
    | X.PaLet(lb,_,_,binds,p) ->
      group (label lb ^^ string "let" ^^ (nest 2 (hardline ^^ nest 2 (string "| " ^^ separate_map (hardline ^^ string "and ") letbind binds))) ^/^ group (string "in" ^/^ plisty p))
    | X.PaCtx(_,p) -> plisty p
    | X.PaInst(lb, x, _) -> group (label lb ^^ string "lifting _")
    | X.PaClause(lb,_,u,v) ->
      group ((label lb ^^ string "clause" ^//^
               group (
                 string "| " ^^
                   nest 2 (separate_map (string ";" ^^ break 1) papty u) ^/^ string "->" ^^ (nest 2 (break 1 ^^ plisty v)))) ^/^ string "end")
    | X.PaCon(lb,_,_,_) -> label lb ^^ string "_"
  and papty = function
    | X.PaPaCon(lb,_,_,p), _ -> label lb ^^ string "_"
  and letbind (lb,_,name,t,p) =
    group (label lb ^^ id name ^^ string " ornament _" ^//^
             (string "patch" ^//^ plisty p))
  and plisty = function
    | [] -> string ""
    | [x] -> pty x
    | ((X.PaLet _, _) as x)::xs ->
      lparen ^^ pty x ^^ rparen ^^ string ";" ^/^ plisty xs
    | x::xs -> pty x ^^ string ";" ^/^ plisty xs
  in
  group
    (string "let " ^^ id name ^^ string " =" ^//^
       (string "lifting of " ^^ id name ^/^
       string "using" ^/^
       string "ornament _"  ^/^
       string "patch" ^//^
       (plisty p)))

 *)
open Orn
let oty =
  let rec ty p ppf t = match t, p with
    | OArrow(t1,t2), false -> fp ppf "%a -> %a" (ty true) t1 (ty false) t2
    | OArrow(t1,t2), true ->  fp ppf "(%a -> %a)" (ty true) t1 (ty false) t2
    | OVar(x),_ -> tyvar ppf x
    | OProduct ts,_ -> fp ppf "(%a)" (seq ", " (ty false)) ts
    | (OIdOrn(c,[]) | OBaseOrn(c, [])),_ -> fp ppf "%a" id c
    | (OIdOrn(c,[t]) | OBaseOrn(c, [t])), _ -> fp ppf "%a %a" (ty true) t id c
    | (OIdOrn(c,ts) | OBaseOrn(c,ts)),_ -> fp ppf "(%a) %a" (seq ", " (ty false)) ts id c
  in
  ty false

let ornvar ppf (x,c,ws) = fp ppf "(%a < %a)" id x oty (OIdOrn(c,ws))
let osc ppf osc =
  fp ppf "∀(%a) ∀%a. %a" (seq " " id) osc.ornscheme_paravars (seq "" ornvar) osc.ornscheme_ornvars oty osc.ornscheme_ornty

let (eml_expr,eml_typ, eml_expr') =
  let open Eml.ExplicitGml in
  let let_type = function
    | IsMeta -> "meta"
    | IsShare -> "share"
    | IsUser -> "user"
  in
  let rec term' ppf = function
    | Absurd -> fp ppf "<absurd>"
    | Var(v) -> fp ppf "%a" id v
    | Let(is_meta,bs,e) -> fp ppf "(let<%s> %a in@\n%a)" (let_type is_meta) (seq " and " letb) bs term e
    | MultiFix(bs, x) -> fp ppf "(fix %a in %a)" (seq " and " fixb) bs id x
    | Match(e, cs, abs) -> fp ppf "(match %a with@\n%a | <absurd> -> %a end)" term e (seq "" clause) cs term abs
    | Con(c,ts,es) -> fp ppf "%a[%a](%a)" id c (seq "," typ) ts (seq "," term) es
    | Tuple(ts,es) -> fp ppf "[%a](%a)" (seq "," typ) ts (seq "," term) es
    | LetTuple(bs,e,f) -> fp ppf "(let (%a) = %a in %a)" (seq "," ltb) bs term e term f
    | Abs(x,ty,e) -> fp ppf "(fun (%a : %a) -> %a)" id x typ ty term e
    | App(e,f) -> fp ppf "(%a %a)" term e term f
    | GAbs(x,e) -> fp ppf "(fun %a #> %a)" id x term e
    | GApp(e,f) -> fp ppf "(%a # %a)" term e term f
    | GAbsT(x, e) -> fp ppf "(fun %a #> %a)" tyvar x term e
    | GAppT(e, ty) -> fp ppf "(%a # %a)" term e typ ty
    | Label(s, e) -> fp ppf "(* %s *) %a" s term e
    | Mark(m, e) -> fp ppf "marked(%a)" term e
  and term ppf (e,_,_) = term' ppf e
  and letb ppf (x,ty,e) = fp ppf "%a : %a = %a" id x typ ty term e
  and fixb ppf (x,(y,ty),e) = fp ppf "%a (%a : %a) -> %a" id x id y typ ty term e
  and clause ppf (c,xs,e) = fp ppf "| %a(%a) -> @[<h>%a@]@\n" id c (seq "," id) xs term e
  and ltb ppf (x,ty) = fp ppf "%a : %a" id x typ ty
  and typ ppf = function
    | TFall(x, ty) -> fp ppf "(∀ %a. %a)" tyvar x typ ty
    | TGFall(x,ty) -> fp ppf "(∀# %a. %a)" tyvar x typ ty
    | TMatch(e,cs) -> fp ppf "(match %a with %a)" term e (seq " | " tclause) cs
    | TCon(c,ts) -> fp ppf "(%a %a)" id c (seq " " typ) ts
    | TVar(v) -> fp ppf "%a" tyvar v
    | TArr(e,f) -> fp ppf "(%a -> %a)" typ e typ f
    | TGArr(x,t1,t2) -> fp ppf "(Π(%a : %a) %a)" id x typ t1 typ t2
    | TAbs(x,t1,t2) -> fp ppf "(λ(%a : %a) #> %a)" id x typ t1 typ t2
    | TApp(t,e) -> fp ppf "(%a # %a)" typ t term e
    | TAbsT(x, t) -> fp ppf "(Λ(%a) #> %a)" tyvar x typ t
    | TAppT(t1,t2) -> fp ppf "(%a # %a)" typ t1 typ t2
    | TProd(ts) -> fp ppf "(%a)" (seq " * " typ) ts
  and tclause ppf (c,xs,t) = fp ppf "%a(%a) -> %a" id c (seq "," id) xs typ t
  in
  (term, typ,term')

let orn_def ppf (odef:orn_def) =
  let { orn_name; orn_args; orn_from; orn_from_skel = Scheme(args_skel, ty_skel); orn_to; orn_clauses; orn_fallback }
    = odef in
  fp ppf "%a = ∀(%a) ∀(%a) %a => %a => %a@\n%a@."
    id_nice orn_name
    (seq " " id_nice) orn_args
    (seq " " id_nice) args_skel
    ty_nice orn_from
    ty_nice ty_skel
    ty_nice orn_to
    (seq "\n" (fun ppf (l, r) -> fp ppf "%a => %a" pat_nice l pat_nice r)) orn_clauses

let impl_term_stderr t =
  PPrint.ToChannel.pretty 1. 60 stderr (piterm Expr t)

let impl_pat_stderr t =
  PPrint.ToChannel.pretty 1. 60 stderr (pipat t)
