(***********************************************************************)
(*                                                                     *)
(*                                OCaml                                *)
(*                                                                     *)
(*            Xavier Leroy, projet Cristal, INRIA Rocquencourt         *)
(*                                                                     *)
(*  Copyright 1996 Institut National de Recherche en Informatique et   *)
(*  en Automatique.  All rights reserved.  This file is distributed    *)
(*  under the terms of the GNU Library General Public License, with    *)
(*  the special exception on linking described in file ../LICENSE.     *)
(*                                                                     *)
(***********************************************************************)

(* $Id$ *)

(* Sets over ordered types *)
#(mode input)

type t = Empty | Node of t * elt * t * int
let invalid_arg = Empty


(* Sets are represented by balanced binary trees (the heights of the
       children differ by at most 2 *)

let height t = match t with
    Empty -> i0
  | Node(_, _, _, h) -> h

(* Creates a new node with left son l, value v and right son r.
       We must have all elements of l < v < all elements of r.
       l and r must be balanced and | height l - height r | <= 2.
       Inline expansion of height for better speed. *)

let create l vb r =
  match vb with
  | Box v ->
  let hl = match l with Empty -> i0 | Node(_,_,_,h) -> h in
  let hr = match r with Empty -> i0 | Node(_,_,_,h) -> h in
  Node(l, v, r, (match gte hl hr with True -> plus hl i1 | False -> plus hr i1))

(* Same as create, but performs one step of rebalancing if necessary.
   Assumes l and r balanced and | height l - height r | <= 3.
   Inline expansion of create for better speed in the most frequent case
   where no rebalancing is required. *)

let bal l vb r =
  match vb with
  | Box v ->
  let hl = match l with Empty -> i0 | Node(_,_,_,h) -> h in
  let hr = match r with Empty -> i0 | Node(_,_,_,h) -> h in
  match gt hl (plus hr i2) with
  | True ->
     begin
       match l with
         Empty -> invalid_arg
       | Node(ll, lv, lr, _) ->
          match gte (height ll) (height lr) with
          | True -> create ll (Box lv) (create lr (Box v) r)
          | False -> begin
              match lr with
                Empty -> invalid_arg
              | Node(lrl, lrv, lrr, _)->
                 create (create ll (Box lv) lrl) (Box lrv) (create lrr (Box v) r)
            end
     end
  | False ->
     begin match gt hr (plus hl i2) with
           | True -> begin
                     match r with
                       Empty -> invalid_arg
                     | Node(rl, rv, rr, _) ->
                        match gte (height rr) (height rl) with
                        | True -> create (create l (Box v) rl) (Box rv) rr
                        | False -> begin
                            match rl with
                              Empty -> invalid_arg
                            | Node(rll, rlv, rlr, _) ->
                               create (create l (Box v) rll) (Box rlv) (create rlr (Box rv) rr)
                          end
                   end
           | False ->  Node(l, v, r, (match gte hl hr with True -> plus hl i1 | False -> plus hr i1))
     end

(* Insertion of one element *)

let rec add xb t =
  match xb with
  | Box x ->
     match t with
       Empty -> Node(Empty, x, Empty, i1)
     | Node(l, v, r, _) ->
        match compare x v with
        | Eq -> t
        | Lt -> bal (add (Box x) l) (Box v) r
        | Gt -> bal l (Box v) (add (Box x) r)

let singleton xb =
  match xb with
  | Box x -> Node(Empty, x, Empty, i1)


(* Smallest and greatest element of a set *)

let rec min_elt t = match t with
    Empty -> None
  | Node(Empty, v, r, _) -> Some v
  | Node(l, v, r, _) -> min_elt l

let rec max_elt t = match t with
    Empty -> None
  | Node(l, v, Empty, _) -> Some v
  | Node(l, v, r, _) -> max_elt r

