type int = Int_is_abstract

type string = String_is_abstract

type bool =
  | True
  | False

val add : (int -> int -> int)

val mul : (int -> int -> int)

val succ : (int -> int)

val equal : ('a -> 'a -> bool)

val i0 : int
val i1 : int
val i2 : int
val i42 : int
val gt : int -> int -> bool
val gte : int -> int -> bool
val plus : int -> int -> int

type 'a option = None | Some of 'a

type 'a box = Box of 'a
type cmp = Eq | Lt | Gt

type elt = Elt
val compare: elt -> elt -> cmp

val fail: 'a


type t = Empty | Node of t * elt * t * int
let invalid_arg = Empty


(* Sets are represented by balanced binary trees (the heights of the
       children differ by at most 2 *)

let height t = match t with
    Empty -> i0
  | Node(_, _, _, h) -> h

(* Creates a new node with left son l, value v and right son r.
       We must have all elements of l < v < all elements of r.
       l and r must be balanced and | height l - height r | <= 2.
       Inline expansion of height for better speed. *)

let create l vb r =
  match vb with
  | Box v ->
  let hl = match l with Empty -> i0 | Node(_,_,_,h) -> h in
  let hr = match r with Empty -> i0 | Node(_,_,_,h) -> h in
  Node(l, v, r, (match gte hl hr with True -> plus hl i1 | False -> plus hr i1))

(* Same as create, but performs one step of rebalancing if necessary.
   Assumes l and r balanced and | height l - height r | <= 3.
   Inline expansion of create for better speed in the most frequent case
   where no rebalancing is required. *)

let bal l vb r =
  match vb with
  | Box v ->
  let hl = match l with Empty -> i0 | Node(_,_,_,h) -> h in
  let hr = match r with Empty -> i0 | Node(_,_,_,h) -> h in
  match gt hl (plus hr i2) with
  | True ->
     begin
       match l with
         Empty -> invalid_arg
       | Node(ll, lv, lr, _) ->
          match gte (height ll) (height lr) with
          | True -> create ll (Box lv) (create lr (Box v) r)
          | False -> begin
              match lr with
                Empty -> invalid_arg
              | Node(lrl, lrv, lrr, _)->
                 create (create ll (Box lv) lrl) (Box lrv) (create lrr (Box v) r)
            end
     end
  | False ->
     begin match gt hr (plus hl i2) with
           | True -> begin
                     match r with
                       Empty -> invalid_arg
                     | Node(rl, rv, rr, _) ->
                        match gte (height rr) (height rl) with
                        | True -> create (create l (Box v) rl) (Box rv) rr
                        | False -> begin
                            match rl with
                              Empty -> invalid_arg
                            | Node(rll, rlv, rlr, _) ->
                               create (create l (Box v) rll) (Box rlv) (create rlr (Box rv) rr)
                          end
                   end
           | False ->  Node(l, v, r, (match gte hl hr with True -> plus hl i1 | False -> plus hr i1))
     end

(* Insertion of one element *)

let rec add xb t =
  match xb with
  | Box x ->
     match t with
       Empty -> Node(Empty, x, Empty, i1)
     | Node(l, v, r, _) ->
        match compare x v with
        | Eq -> t
        | Lt -> bal (add (Box x) l) (Box v) r
        | Gt -> bal l (Box v) (add (Box x) r)

let singleton xb =
  match xb with
  | Box x -> Node(Empty, x, Empty, i1)


(* Smallest and greatest element of a set *)

let rec min_elt t = match t with
    Empty -> None
  | Node(Empty, v, r, _) -> Some v
  | Node(l, v, r, _) -> min_elt l

let rec max_elt t = match t with
    Empty -> None
  | Node(l, v, Empty, _) -> Some v
  | Node(l, v, r, _) -> max_elt r

  (** Lifting to maps **)
  type 'a map = MEmpty | MNode of 'a map * elt * 'a * 'a map * int

  type ornament 'a setmap : t => 'a map with
       | Empty => MEmpty
       | Node(l,v,r,n) => MNode(l,v,_,r,n) when l r : 'a setmap

  type ornament ('a,'new) box2 : 'a box => ('a * 'new) box with
       | Box a => Box (a,_)

  type ornament ('a, 'b) option2 : 'a option => ('a * 'b) option with
       | None => None
       | Some u => Some (u, _)

  let minvalid_arg = lifting invalid_arg : _ setmap

  let mheight = lifting height : _ setmap -> _

  let mcreate = lifting create : _ setmap -> _ box2 -> _ setmap -> _ setmap with
                |  (MNode(l, _c, #a, _d, _a)) => #a <- (match vb with Box (_,x) -> x)

  let mbal = lifting bal : _ setmap -> _ box2 -> _ setmap -> _ setmap with
    | ornament * <- setmap, box2
    | (mcreate l (Box(_v, (#a)))) => #a <- (match vb with Box(_,x) -> x)
    | (mcreate (mcreate l (Box (_a, _b)) _l1) (Box (_v1, #a)) (mcreate _r1 (Box (_v, _x1)) _r)) => #a <- (match r with MNode(MNode(_,_,x,_,_),_,_,_,_) -> x)
    | (mcreate (mcreate l (Box (_a, _b)) _l1) (Box (_v1, _x1)) (mcreate _r1 (Box (_v, #a)) _r)) => #a <- (match r with MNode(_,_,x,_,_) -> x)
    | (mcreate (mcreate _l (Box (_v, #a)) _l1)
                      (Box (_v1,  _b))
                      (mcreate _r1 (Box (_a, _c)) r)) => #a <- (match l with MNode(_,_,x,_,_) -> x)
    | (mcreate _l (Box (_v, #a))
                  (mcreate _r1 (Box (_a, _b)) r)) => #a <- (match l with MNode(_,_,x,_,_) -> x)
    | (mcreate _l (Box (_v, _b))
                  (mcreate _r1 (Box (_a, #a)) r)) => #a <- (match vb with Box(_,x) -> x)
    | (mcreate (mcreate _l (Box (_v, _c)) _l1)
                      (Box (_v1,  _b))
                      (mcreate _r1 (Box (_a, #a)) r)) => #a <- (match vb with Box(_,x) -> x)
    | (mcreate (mcreate _l (Box (_v, _c)) _l1)
                      (Box (_v1,  #a))
                      (mcreate _r1 (Box (_a, _b)) r)) => #a <- (match l with MNode(_,_,_,MNode(_,_,x,_,_),_) -> x)

let madd = lifting add : _ box2 -> _ setmap -> _ setmap with
  | (MNode(_a, _b, #a, _c, _d)) => #a <- (match xb with Box(_, x) -> x)
  | (mbal _l (Box(_v, #a))) => #a <- (match t with MNode(_, _, x, _, _) -> x)
  | (_madd (Box (_a, #a)) _l) => #a <- (match xb with Box(_, x) -> x)


  let msingleton = lifting singleton : _ box2 -> _ setmap with
    | #a <- (match xb with Box (_,x) -> x)


  let mmin_binding = lifting min_elt : _ setmap -> _ option2 with
    | #a <-  (match t with MNode(_,_,x,_,_) -> x)

  let mmax_binding = lifting max_elt : _ setmap -> _ option2 with
    | #a <- (match t with MNode(_,_,x,_,_) -> x)
