open Str

let file_to_lines filename =
  let fd = open_in filename in
  let rec input all =
    try input (input_line fd :: all)
    with End_of_file -> close_in fd; all
  in
  let all = input [] in
  List.mapi (fun i x ->  (i+2), x) (List.rev all)

let begin_program =
  Str.regexp "^ *\\\\begin{program}\\(\\[[^]]*\\]\\)?\\({[^}]*}\\)?"
let end_program = Str.regexp "^ *\\\\end{program}"

type put = Library | Input | Lifting | Inout | Skip | Output
type line = Begin of put | End | Line of string

let classify ((i : int), str) : int * line =
  try
    let _pos = search_forward begin_program str 0 in
    let matched n str =
      try let s = matched_group n str in String.sub s 1 (String.length s - 2)
      with Not_found -> "" in
    let bracket = matched 1 str in
    let braces = matched 2 str in
    let put =
      match if bracket = "" then braces else bracket  with
      | "library" -> Library
      | "input" -> Input
      | "lifting" -> Lifting
      | "output" -> Output
      | "inout" -> Inout
      | "postin" -> Skip
      | "postout" -> Skip
      | _ -> Skip in
    i, Begin put
  with Not_found ->
    try
      let _pos = search_forward end_program str 0 in
      i, End
    with Not_found ->
      i, Line str

exception Unclosed of int * int
let process_lines texname lines library iput lifting oput standalone =
  let out fd str =
    output_string fd str; output_char fd '\n' in
  let output is_mode put str =
    begin
      match standalone, put with
      | Some o, (Library | Input | Lifting | Inout) when not is_mode ->
        out o str
      | _ -> ()
    end;
    match put with
    | Library -> out library str
    | Input -> out iput str
    | Lifting -> out lifting str
    | Output -> out oput str
    | Inout -> out lifting str; out oput str
    | Skip -> ()
  in

  let rec skip beg  = function
    | (i, Line str) :: rest ->
        skip beg rest
    | (i, End) :: _ ->
        raise (Unclosed (i, beg))
    | (i, Begin put) :: rest ->
        begin match put with
        | Input | Lifting | Library ->
            output true put (Printf.sprintf "\n# %d \"%s\"" i texname);
        | Inout ->
            output true Lifting (Printf.sprintf "\n# %d \"%s\"" i texname);
        | _ -> ()
        end;
        copy i put rest
    | [] -> ()
  and copy beg put = function
    | (i, Line str) :: rest ->
        output false put str;
        copy beg put rest
    | (i, End) :: rest ->
        skip i rest
    | (i, Begin put) :: rest ->
        raise (Unclosed (i, beg))
    | [] ->
        raise (Unclosed (10000, beg)) in
  skip 0  lines


let process src =
  let basename = Filename.chop_extension src in
  let texname = basename ^ ".tex" in
  if not (src = texname) then failwith (src ^ "does not end with .tex");
  Printf.eprintf "Processing %s\n%!" texname;
  let strings = file_to_lines texname in
  let lines = List.map classify strings in
  let library = open_out (basename ^ ".lib.ml") in
  let iput = open_out (basename ^ ".in.ml") in
  let lifting = open_out (basename ^ ".lif.ml") in
  let oput = open_out (basename ^ ".ref.ml") in
  let standalone = open_out (basename ^ ".ml") in
  let () =
    try process_lines texname lines library iput lifting oput (Some standalone)
    with Unclosed (b,e) ->
      Printf.printf "Unbalance (%s, %d, %d)" src b e
  in
  close_out iput;
  close_out oput;
  close_out library;
  close_out lifting;
  ()

let main () =
  let files = List.tl (Array.to_list Sys.argv) in
  List.iter process files

let () = main()
